from typing import List

from fastapi import APIRouter, Depends, HTTPException, Request, status
from fastapi.responses import JSONResponse
from fastapi_users.exceptions import InvalidPasswordException, UserAlreadyExists
from sqlalchemy.orm import joinedload

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.user_field import get_user_fields
from fief.dependencies.users import (
    UserManager,
    get_admin_user_update,
    get_paginated_users,
    get_user_by_id_or_404,
    get_user_create_internal,
    get_user_manager_from_create_user_internal,
    get_user_manager_from_user,
)
from fief.dependencies.workspace_managers import get_user_manager as get_user_db_manager
from fief.errors import APIErrorCode
from fief.managers import UserManager as UserDBManager
from fief.models import User, UserField
from fief.schemas.generics import PaginatedResults

router = APIRouter(dependencies=[Depends(is_authenticated_admin)])


@router.get(
    "/", name="users:list", response_model=PaginatedResults[schemas.user.UserRead]
)
async def list_users(
    paginated_users: PaginatedObjects[User] = Depends(get_paginated_users),
) -> PaginatedResults[schemas.user.UserRead]:
    users, count = paginated_users
    return PaginatedResults(
        count=count,
        results=[schemas.user.UserRead.from_orm(user) for user in users],
    )


@router.post(
    "/",
    name="users:create",
    response_model=schemas.user.UserRead,
    status_code=status.HTTP_201_CREATED,
)
async def create_user(
    request: Request,
    user_create: schemas.user.UserCreateInternal = Depends(get_user_create_internal),
    user_fields: List[UserField] = Depends(get_user_fields),
    user_manager: UserManager = Depends(get_user_manager_from_create_user_internal),
    user_db_manager: UserDBManager = Depends(get_user_db_manager),
):
    try:
        created_user = await user_manager.create_with_fields(
            user_create, user_fields=user_fields, request=request
        )
    except UserAlreadyExists as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_CREATE_ALREADY_EXISTS,
        ) from e
    except InvalidPasswordException as e:
        # Build a JSON response manually to fine-tune the response structure
        return JSONResponse(
            content={
                "detail": APIErrorCode.USER_CREATE_INVALID_PASSWORD,
                "reason": e.reason,
            },
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    user = await user_db_manager.get_by_id(created_user.id, (joinedload(User.tenant),))

    return schemas.user.UserRead.from_orm(user)


@router.patch("/{id:uuid}", name="users:update", response_model=schemas.user.UserRead)
async def update_user(
    request: Request,
    user_update: schemas.user.UserUpdate = Depends(get_admin_user_update),
    user: User = Depends(get_user_by_id_or_404),
    user_fields: List[UserField] = Depends(get_user_fields),
    user_manager: UserManager = Depends(get_user_manager_from_user),
):
    try:
        user = await user_manager.update_with_fields(
            user_update,
            user,
            user_fields=user_fields,
            safe=False,
            request=request,
        )
    except UserAlreadyExists as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_UPDATE_EMAIL_ALREADY_EXISTS,
        ) from e
    except InvalidPasswordException as e:
        # Build a JSON response manually to fine-tune the response structure
        return JSONResponse(
            content={
                "detail": APIErrorCode.USER_UPDATE_INVALID_PASSWORD,
                "reason": e.reason,
            },
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    return schemas.user.UserRead.from_orm(user)



from typing import Optional

from pydantic import BaseModel, EmailStr, Field


class ForgotPasswordRequest(BaseModel):
    email: EmailStr


class ResetPasswordError(BaseModel):
    error: str = Field(..., regex="missing_token|invalid_token|invalid_password")
    error_description: Optional[str] = None
    error_uri: Optional[str] = None

    @classmethod
    def get_missing_token(cls, error_description: Optional[str] = None):
        return cls(error="missing_token", error_description=error_description)

    @classmethod
    def get_invalid_token(cls, error_description: Optional[str] = None):
        return cls(error="invalid_token", error_description=error_description)

    @classmethod
    def get_invalid_password(cls, error_description: Optional[str] = None):
        return cls(error="invalid_password", error_description=error_description)


class ResetPasswordRequest(BaseModel):
    token: str
    password: str



from pydantic import UUID4
from sqlalchemy import Column, ForeignKey
from sqlalchemy.orm import relationship

from fief.models.base import MainBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.workspace import Workspace


class WorkspaceUser(UUIDModel, CreatedUpdatedAt, MainBase):
    __tablename__ = "workspace_users"

    workspace_id: UUID4 = Column(GUID, ForeignKey(Workspace.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    user_id: UUID4 = Column(GUID, nullable=False)

    workspace: Workspace = relationship(
        "Workspace", back_populates="workspace_users", lazy="joined"
    )

    def __repr__(self) -> str:
        return f"WorkspaceUser(id={self.id}, workspace_id={self.workspace_id}, user_id={self.user_id})"



from pydantic import UUID4, BaseModel, SecretStr

from fief.schemas.generics import CreatedUpdatedAt, UUIDSchema


class AdminAPIKeyCreate(BaseModel):
    name: str


class BaseAdminApiKey(UUIDSchema, CreatedUpdatedAt):
    name: str
    workspace_id: UUID4


class AdminAPIKeyCreateResponse(BaseAdminApiKey):
    token: str


class AdminAPIKey(BaseAdminApiKey):
    token: SecretStr



from fastapi import Depends

from fief.dependencies.main_managers import (
    get_workspace_manager,
    get_workspace_user_manager,
)
from fief.dependencies.workspace_db import get_workspace_db
from fief.managers import WorkspaceManager, WorkspaceUserManager
from fief.services.workspace_creation import WorkspaceCreation
from fief.services.workspace_db import WorkspaceDatabase


async def get_workspace_creation(
    workspace_manager: WorkspaceManager = Depends(get_workspace_manager),
    workspace_user_manager: WorkspaceUserManager = Depends(get_workspace_user_manager),
    workspace_db: WorkspaceDatabase = Depends(get_workspace_db),
) -> WorkspaceCreation:
    return WorkspaceCreation(workspace_manager, workspace_user_manager, workspace_db)



from typing import Any, Dict, List, Type

from fastapi import APIRouter, Depends, Request, status
from fastapi.responses import RedirectResponse
from fastapi_users.exceptions import InvalidPasswordException, UserAlreadyExists

from fief.apps.auth.templates import templates
from fief.csrf import check_csrf
from fief.dependencies.auth import get_login_session
from fief.dependencies.authentication_flow import get_authentication_flow
from fief.dependencies.form import get_form_data_dict
from fief.dependencies.locale import Translations, get_gettext, get_translations
from fief.dependencies.register import get_register_context, get_user_create
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.user_field import get_user_create_internal_model, get_user_fields
from fief.dependencies.users import UserManager, get_user_manager
from fief.exceptions import RegisterException
from fief.models import Tenant, UserField
from fief.schemas.register import RegisterError
from fief.schemas.user import UF, UserCreate, UserCreateInternal
from fief.services.authentication_flow import AuthenticationFlow

router = APIRouter(dependencies=[Depends(check_csrf), Depends(get_translations)])


@router.get("/register", name="register:get", dependencies=[Depends(get_login_session)])
async def get_register(
    request: Request,
    register_context: Dict[str, Any] = Depends(get_register_context),
    translations: Translations = Depends(get_translations),
):
    return templates.LocaleTemplateResponse(
        "register.html",
        {
            "request": request,
            **register_context,
        },
        translations=translations,
    )


@router.post(
    "/register", name="register:post", dependencies=[Depends(get_login_session)]
)
async def post_register(
    request: Request,
    user: UserCreate[UF] = Depends(get_user_create),
    form_data_dict: Dict[str, Any] = Depends(get_form_data_dict),
    user_create_internal_model: Type[UserCreateInternal[UF]] = Depends(
        get_user_create_internal_model
    ),
    user_manager: UserManager = Depends(get_user_manager),
    user_fields: List[UserField] = Depends(get_user_fields),
    register_context: Dict[str, Any] = Depends(get_register_context),
    tenant: Tenant = Depends(get_current_tenant),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    _=Depends(get_gettext),
):
    try:
        user_create = user_create_internal_model(
            **user.dict(), fields=user.fields, tenant_id=tenant.id
        )
        created_user = await user_manager.create_with_fields(
            user_create,
            user_fields=user_fields,
            safe=True,
            request=request,
        )
    except UserAlreadyExists as e:
        raise RegisterException(
            RegisterError.get_user_already_exists(
                _("A user with the same email address already exists.")
            ),
            context=register_context,
            form_data=form_data_dict,
        ) from e
    except InvalidPasswordException as e:
        raise RegisterException(
            RegisterError.get_invalid_password(e.reason),
            context=register_context,
            form_data=form_data_dict,
        ) from e

    response = RedirectResponse(
        tenant.url_for(request, "auth:consent.get"),
        status_code=status.HTTP_302_FOUND,
    )
    response = await authentication_flow.create_session_token(response, created_user.id)

    return response



from fastapi import Depends
from sqlalchemy import select

from fief.dependencies.admin_session import get_admin_session_token
from fief.dependencies.main_managers import get_workspace_manager
from fief.dependencies.pagination import (
    Ordering,
    PaginatedObjects,
    Pagination,
    get_ordering,
    get_paginated_objects,
    get_pagination,
)
from fief.managers import WorkspaceManager
from fief.models import AdminSessionToken, Workspace, WorkspaceUser


async def get_paginated_workspaces(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(get_ordering),
    manager: WorkspaceManager = Depends(get_workspace_manager),
    admin_session_token: AdminSessionToken = Depends(get_admin_session_token),
) -> PaginatedObjects[Workspace]:
    statement = (
        select(Workspace)
        .join(Workspace.workspace_users)
        .where(WorkspaceUser.user_id == admin_session_token.user_id)
    )
    return await get_paginated_objects(statement, pagination, ordering, manager)



from typing import Optional

from fastapi import APIRouter, Depends, Query, Request, status
from fastapi.responses import RedirectResponse
from fastapi_users.exceptions import (
    InvalidPasswordException,
    InvalidResetPasswordToken,
    UserInactive,
    UserNotExists,
)

from fief.apps.auth.templates import templates
from fief.csrf import check_csrf
from fief.dependencies.auth import get_optional_login_session
from fief.dependencies.locale import Translations, get_gettext, get_translations
from fief.dependencies.reset import (
    get_forgot_password_request,
    get_reset_password_request,
)
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.users import UserManager, get_user_manager
from fief.exceptions import ResetPasswordException
from fief.models import LoginSession, Tenant
from fief.schemas.reset import (
    ForgotPasswordRequest,
    ResetPasswordError,
    ResetPasswordRequest,
)

router = APIRouter(dependencies=[Depends(check_csrf), Depends(get_translations)])


@router.get("/forgot", name="reset:forgot.get")
async def get_forgot_password(
    request: Request,
    tenant: Tenant = Depends(get_current_tenant),
    translations: Translations = Depends(get_translations),
):
    return templates.LocaleTemplateResponse(
        "forgot_password.html",
        {"request": request, "tenant": tenant},
        translations=translations,
    )


@router.post("/forgot", name="reset:forgot.post")
async def post_forgot_password(
    request: Request,
    forgot_password_request: ForgotPasswordRequest = Depends(
        get_forgot_password_request
    ),
    user_manager: UserManager = Depends(get_user_manager),
    tenant: Tenant = Depends(get_current_tenant),
    translations: Translations = Depends(get_translations),
):
    try:
        user = await user_manager.get_by_email(forgot_password_request.email)
        await user_manager.forgot_password(user, request)
    except (UserNotExists, UserInactive):
        pass

    success = translations.gettext(
        "Check your inbox! If an account associated with this email address exists in our system, you'll receive a link to reset your password."
    )

    return templates.LocaleTemplateResponse(
        "forgot_password.html",
        {"request": request, "tenant": tenant, "success": success},
        translations=translations,
    )


@router.get("/reset", name="reset:reset.get")
async def get_reset_password(
    request: Request,
    token: Optional[str] = Query(None),
    tenant: Tenant = Depends(get_current_tenant),
    translations: Translations = Depends(get_translations),
    _=Depends(get_gettext),
):
    if token is None:
        raise ResetPasswordException(
            ResetPasswordError.get_missing_token(
                _("The reset password token is missing.")
            ),
            tenant=tenant,
            fatal=True,
        )

    return templates.LocaleTemplateResponse(
        "reset_password.html",
        {"request": request, "tenant": tenant, "token": token},
        translations=translations,
    )


@router.post("/reset", name="reset:reset.post")
async def post_reset_password(
    request: Request,
    reset_password_request: ResetPasswordRequest = Depends(get_reset_password_request),
    user_manager: UserManager = Depends(get_user_manager),
    login_session: Optional[LoginSession] = Depends(get_optional_login_session),
    tenant: Tenant = Depends(get_current_tenant),
    translations: Translations = Depends(get_translations),
    _=Depends(get_gettext),
):
    try:
        await user_manager.reset_password(
            reset_password_request.token, reset_password_request.password, request
        )
    except (InvalidResetPasswordToken, UserNotExists, UserInactive) as e:
        raise ResetPasswordException(
            ResetPasswordError.get_invalid_token(
                _("The reset password token is invalid or expired.")
            ),
            tenant=tenant,
            fatal=True,
        ) from e
    except InvalidPasswordException as e:
        raise ResetPasswordException(
            ResetPasswordError.get_invalid_password(e.reason),
            form_data=await request.form(),
            tenant=tenant,
        ) from e

    if login_session is not None:
        redirection = tenant.url_for(request, "auth:login.get")
        return RedirectResponse(url=redirection, status_code=status.HTTP_302_FOUND)

    return templates.LocaleTemplateResponse(
        "reset_password.html",
        {
            "request": request,
            "tenant": tenant,
            "success": _("Your password has been changed!"),
        },
        translations=translations,
    )



from typing import Optional

from fastapi import Depends, HTTPException, status

from fief.dependencies.admin_api_key import get_optional_admin_api_key
from fief.dependencies.admin_session import get_optional_admin_session_token
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.main_managers import get_workspace_user_manager
from fief.managers.workspace_user import WorkspaceUserManager
from fief.models import AdminAPIKey, AdminSessionToken, Workspace


async def is_authenticated_admin(
    session_token: Optional[AdminSessionToken] = Depends(
        get_optional_admin_session_token
    ),
    admin_api_key: Optional[AdminAPIKey] = Depends(get_optional_admin_api_key),
    current_workspace: Workspace = Depends(get_current_workspace),
    workspace_user_manager: WorkspaceUserManager = Depends(get_workspace_user_manager),
):
    if session_token is None and admin_api_key is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    if session_token is not None:
        workspace_user = await workspace_user_manager.get_by_workspace_and_user(
            current_workspace.id, session_token.user_id
        )
        if workspace_user is None:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)
    elif admin_api_key is not None:
        if admin_api_key.workspace_id != current_workspace.id:
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)



from enum import Enum
from typing import Dict, Type

from fief.services.email.base import EmailError, EmailProvider, SendEmailError
from fief.services.email.null import Null
from fief.services.email.postmark import Postmark


class AvailableEmailProvider(str, Enum):
    NULL = "NULL"
    POSTMARK = "POSTMARK"


EMAIL_PROVIDERS: Dict[AvailableEmailProvider, Type[EmailProvider]] = {
    AvailableEmailProvider.NULL: Null,
    AvailableEmailProvider.POSTMARK: Postmark,
}

__all__ = [
    "AvailableEmailProvider",
    "EMAIL_PROVIDERS",
    "EmailError",
    "EmailProvider",
    "SendEmailError",
]



from typing import Any

from fastapi import Request
from jwcrypto import jwk
from sqlalchemy import Boolean, Column, String, Text

from fief.crypto.jwk import generate_signature_jwk_string, load_jwk
from fief.models.base import WorkspaceBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel


class Tenant(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "tenants"

    name: str = Column(String(length=255), nullable=False)
    slug: str = Column(String(length=255), nullable=False, unique=True)
    default: bool = Column(Boolean, default=False, nullable=False)
    sign_jwk: str = Column(Text, nullable=False, default=generate_signature_jwk_string)

    def __repr__(self) -> str:
        return f"Tenant(id={self.id}, name={self.name}, slug={self.slug}, default={self.default})"

    def get_sign_jwk(self) -> jwk.JWK:
        return load_jwk(self.sign_jwk)

    def get_host(self, domain: str) -> str:
        host = f"https://{domain}"
        if not self.default:
            host += f"/{self.slug}"
        return host

    def url_for(self, request: Request, name: str, **path_params: Any) -> str:
        if not self.default:
            path_params["tenant_slug"] = self.slug
        return request.url_for(name, **path_params)



from typing import List

from pydantic import UUID4
from sqlalchemy import JSON, Column, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import UniqueConstraint

from fief.models.base import WorkspaceBase
from fief.models.client import Client
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.user import User


class Grant(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "grants"
    __table_args__ = (UniqueConstraint("user_id", "client_id"),)

    scope: List[str] = Column(JSON, nullable=False, default=list)

    user_id: UUID4 = Column(GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    user: User = relationship("User")

    client_id: UUID4 = Column(GUID, ForeignKey(Client.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    client: Client = relationship("Client", lazy="joined")



from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base

from fief.settings import settings

MainBase = declarative_base()


class WorkspaceMeta(DeclarativeMeta):
    def __init__(cls, name, bases, dict_):
        try:
            cls.__tablename__ = dict_[
                "__tablename__"
            ] = f"{settings.workspace_table_prefix}{dict_['__tablename__']}"
        except KeyError:
            pass
        return super().__init__(name, bases, dict_)


WorkspaceBase = declarative_base(metaclass=WorkspaceMeta)



from datetime import datetime
from typing import AsyncGenerator, List, Optional, TypedDict

from fastapi import Depends, Form
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from fastapi_users.exceptions import UserNotExists
from pydantic import UUID4

from fief.crypto.code_challenge import verify_code_verifier
from fief.crypto.token import get_token_hash
from fief.dependencies.users import UserManager, get_user_manager
from fief.dependencies.workspace_managers import (
    get_authorization_code_manager,
    get_client_manager,
    get_refresh_token_manager,
)
from fief.exceptions import TokenRequestException
from fief.managers import AuthorizationCodeManager, ClientManager, RefreshTokenManager
from fief.models import Client, ClientType, User
from fief.schemas.auth import TokenError

ClientSecretBasicScheme = HTTPBasic(scheme_name="client_secret_basic", auto_error=False)


class GrantRequest(TypedDict):
    user_id: UUID4
    scope: List[str]
    authenticated_at: datetime
    nonce: Optional[str]
    c_hash: Optional[str]
    client: Client


async def get_grant_type(grant_type: Optional[str] = Form(None)) -> str:
    if grant_type is None:
        raise TokenRequestException(TokenError.get_invalid_request())

    return grant_type


async def authenticate_client_secret_basic(
    credentials: Optional[HTTPBasicCredentials] = Depends(ClientSecretBasicScheme),
    client_manager: ClientManager = Depends(get_client_manager),
) -> Optional[Client]:
    if credentials is None:
        return None

    return await client_manager.get_by_client_id_and_secret(
        credentials.username, credentials.password
    )


async def authenticate_client_secret_post(
    client_id: Optional[str] = Form(None),
    client_secret: Optional[str] = Form(None),
    client_manager: ClientManager = Depends(get_client_manager),
) -> Optional[Client]:
    if client_id is None or client_secret is None:
        return None

    return await client_manager.get_by_client_id_and_secret(client_id, client_secret)


async def authenticate_none(
    client_id: Optional[str] = Form(None),
    client_manager: ClientManager = Depends(get_client_manager),
    grant_type: str = Depends(get_grant_type),
    code_verifier: Optional[str] = Form(None),
) -> Optional[Client]:
    if client_id is None:
        return None

    client = await client_manager.get_by_client_id(client_id)

    if (
        client is None
        or client.client_type != ClientType.PUBLIC
        # Enforce PKCE for authorization_code grant
        or (grant_type == "authorization_code" and code_verifier is None)
    ):
        return None

    return client


async def authenticate_client(
    client_secret_basic: Optional[Client] = Depends(authenticate_client_secret_basic),
    client_secret_post: Optional[Client] = Depends(authenticate_client_secret_post),
    client_none: Optional[Client] = Depends(authenticate_none),
) -> Client:
    if client_secret_basic is not None:
        return client_secret_basic

    if client_secret_post is not None:
        return client_secret_post

    if client_none is not None:
        return client_none

    raise TokenRequestException(TokenError.get_invalid_client())


async def validate_grant_request(
    code: Optional[str] = Form(None),
    code_verifier: Optional[str] = Form(None),
    redirect_uri: Optional[str] = Form(None),
    refresh_token_token: Optional[str] = Form(None, alias="refresh_token"),
    scope: Optional[str] = Form(None),
    grant_type: str = Depends(get_grant_type),
    client: Client = Depends(authenticate_client),
    authorization_code_manager: AuthorizationCodeManager = Depends(
        get_authorization_code_manager
    ),
    refresh_token_manager: RefreshTokenManager = Depends(get_refresh_token_manager),
) -> AsyncGenerator[GrantRequest, None]:
    if grant_type == "authorization_code":
        if code is None:
            raise TokenRequestException(TokenError.get_invalid_request())

        if redirect_uri is None:
            raise TokenRequestException(TokenError.get_invalid_request())

        code_hash = get_token_hash(code)
        authorization_code = await authorization_code_manager.get_valid_by_code(
            code_hash
        )
        if authorization_code is None:
            raise TokenRequestException(TokenError.get_invalid_grant())

        if authorization_code.client.id != client.id:
            raise TokenRequestException(TokenError.get_invalid_grant())

        if authorization_code.redirect_uri != redirect_uri:
            raise TokenRequestException(TokenError.get_invalid_grant())

        code_challenge_tuple = authorization_code.get_code_challenge_tuple()
        if code_challenge_tuple is not None:
            if code_verifier is None or not verify_code_verifier(
                code_verifier, code_challenge_tuple[0], code_challenge_tuple[1]
            ):
                raise TokenRequestException(TokenError.get_invalid_grant())
        elif code_verifier is not None:
            raise TokenRequestException(TokenError.get_invalid_grant())

        yield {
            "user_id": authorization_code.user_id,
            "scope": authorization_code.scope,
            "authenticated_at": authorization_code.authenticated_at,
            "nonce": authorization_code.nonce,
            "c_hash": authorization_code.c_hash,
            "client": client,
        }

        await authorization_code_manager.delete(authorization_code)
        return
    elif grant_type == "refresh_token":
        if refresh_token_token is None:
            raise TokenRequestException(TokenError.get_invalid_request())

        token_hash = get_token_hash(refresh_token_token)
        refresh_token = await refresh_token_manager.get_by_token(token_hash)

        if refresh_token is None:
            raise TokenRequestException(TokenError.get_invalid_grant())

        if refresh_token.client.id != client.id:
            raise TokenRequestException(TokenError.get_invalid_grant())

        new_scope = scope.split() if scope is not None else refresh_token.scope
        if not set(new_scope).issubset(set(refresh_token.scope)):
            raise TokenRequestException(TokenError.get_invalid_scope())

        yield {
            "user_id": refresh_token.user_id,
            "scope": new_scope,
            "authenticated_at": refresh_token.authenticated_at,
            "nonce": None,
            "c_hash": None,
            "client": client,
        }

        await refresh_token_manager.delete(refresh_token)
        return

    raise TokenRequestException(TokenError.get_unsupported_grant_type())


async def get_user_from_grant_request(
    grant_request: GrantRequest = Depends(validate_grant_request),
    user_manager: UserManager = Depends(get_user_manager),
) -> User:
    try:
        return await user_manager.get(grant_request["user_id"])
    except UserNotExists as e:
        raise TokenRequestException(TokenError.get_invalid_grant()) from e



import enum
import secrets
from typing import List, Optional

from jwcrypto import jwk
from pydantic import UUID4
from sqlalchemy import Boolean, Column, Enum, ForeignKey, String, Text
from sqlalchemy.orm import relationship
from sqlalchemy.sql.sqltypes import JSON

from fief.crypto.jwk import load_jwk
from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.tenant import Tenant


def get_default_redirect_uris() -> List[str]:
    return ["http://localhost:8000/docs/oauth2-redirect"]


class ClientType(str, enum.Enum):
    PUBLIC = "public"
    CONFIDENTIAL = "confidential"


class Client(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "clients"

    name: str = Column(String(length=255), nullable=False)
    first_party: bool = Column(Boolean, nullable=False, default=False)

    client_type: ClientType = Column(
        Enum(ClientType, values_callable=lambda x: [e.value for e in x]),
        nullable=False,
        default=ClientType.CONFIDENTIAL,
        server_default=ClientType.CONFIDENTIAL,
    )
    client_id: str = Column(
        String(length=255), default=secrets.token_urlsafe, nullable=False, index=True
    )
    client_secret: str = Column(
        String(length=255), default=secrets.token_urlsafe, nullable=False, index=True
    )
    redirect_uris: List[str] = Column(
        JSON, nullable=False, default=get_default_redirect_uris
    )
    encrypt_jwk: str = Column(Text, nullable=True)

    tenant_id: UUID4 = Column(GUID, ForeignKey(Tenant.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    tenant: Tenant = relationship("Tenant", lazy="joined")

    def __repr__(self) -> str:
        return f"Client(id={self.id}, name={self.name}, client_id={self.client_id})"

    def get_encrypt_jwk(self) -> Optional[jwk.JWK]:
        if self.encrypt_jwk is None:
            return None
        return load_jwk(self.encrypt_jwk)



from typing import Optional

from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import Client


class ClientManager(BaseManager[Client], UUIDManagerMixin[Client]):
    model = Client

    async def get_by_client_id(self, client_id: str) -> Optional[Client]:
        statement = select(Client).where(Client.client_id == client_id)
        return await self.get_one_or_none(statement)

    async def get_by_client_id_and_secret(
        self, client_id: str, client_secret: str
    ) -> Optional[Client]:
        statement = select(Client).where(
            Client.client_id == client_id, Client.client_secret == client_secret
        )
        return await self.get_one_or_none(statement)



from datetime import datetime, timezone
from typing import Optional

from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import RefreshToken


class RefreshTokenManager(BaseManager[RefreshToken], UUIDManagerMixin[RefreshToken]):
    model = RefreshToken

    async def get_by_token(
        self, token: str, *, fresh: bool = True
    ) -> Optional[RefreshToken]:
        statement = select(RefreshToken).where(RefreshToken.token == token)
        if fresh:
            statement = statement.where(
                RefreshToken.expires_at > datetime.now(timezone.utc)
            )

        return await self.get_one_or_none(statement)



from fastapi import APIRouter, Depends, HTTPException, status

from fief.db.types import create_database_url
from fief.dependencies.admin_session import get_admin_session_token
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.workspace import get_paginated_workspaces
from fief.dependencies.workspace_creation import get_workspace_creation
from fief.dependencies.workspace_db import get_workspace_db
from fief.models import AdminSessionToken, Workspace
from fief.schemas.generics import PaginatedResults
from fief.schemas.workspace import (
    WorkspaceCheckConnection,
    WorkspaceCreate,
    WorkspacePublic,
)
from fief.services.workspace_creation import WorkspaceCreation
from fief.services.workspace_db import (
    WorkspaceDatabase,
    WorkspaceDatabaseConnectionError,
)

router = APIRouter()


@router.get(
    "/", name="workspaces:list", dependencies=[Depends(get_admin_session_token)]
)
async def list_workspaces(
    paginated_workspaces: PaginatedObjects[Workspace] = Depends(
        get_paginated_workspaces
    ),
) -> PaginatedResults[WorkspacePublic]:
    workspaces, count = paginated_workspaces
    return PaginatedResults(
        count=count,
        results=[WorkspacePublic.from_orm(workspace) for workspace in workspaces],
    )


@router.post(
    "/check-connection",
    name="workspaces:check_connection",
    status_code=status.HTTP_200_OK,
    dependencies=[Depends(get_admin_session_token)],
)
async def check_connection(
    workspace_check_connection: WorkspaceCheckConnection,
    workspace_db: WorkspaceDatabase = Depends(get_workspace_db),
):
    url = create_database_url(
        workspace_check_connection.database_type,
        asyncio=False,
        username=workspace_check_connection.database_username,
        password=workspace_check_connection.database_password,
        host=workspace_check_connection.database_host,
        port=workspace_check_connection.database_port,
        database=workspace_check_connection.database_name,
    )
    valid, message = workspace_db.check_connection(url)

    if not valid:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=message)


@router.post("/", name="workspaces:create", status_code=status.HTTP_201_CREATED)
async def create_workspace(
    workspace_create: WorkspaceCreate,
    workspace_creation: WorkspaceCreation = Depends(get_workspace_creation),
    admin_session_token: AdminSessionToken = Depends(get_admin_session_token),
) -> WorkspacePublic:
    try:
        workspace = await workspace_creation.create(
            workspace_create, user_id=admin_session_token.user_id
        )
    except WorkspaceDatabaseConnectionError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=e.message,
        ) from e

    return WorkspacePublic.from_orm(workspace)



from typing import Optional, Tuple

from fief.services.email.base import EmailProvider


class Null(EmailProvider):
    def send_email(
        self,
        *,
        sender: Tuple[str, Optional[str]],
        recipient: Tuple[str, Optional[str]],
        subject: str,
        html: Optional[str] = None,
        text: Optional[str] = None,
    ):
        return



import operator
import re
from typing import List, Tuple

from fastapi import Depends, Header, Request

from fief.locale import FALLBACK, Translations, get_preferred_translations

ACCEPT_LANGUAGE_REGEX = re.compile(
    r"""
    (                       # media-range capturing-parenthesis
      [^\s;,]+              # type/subtype
      (?:[ \t]*;[ \t]*      # ";"
        (?:                 # parameter non-capturing-parenthesis
          [^\s;,q][^\s;,]*  # token that doesn't start with "q"
        |                   # or
          q[^\s;,=][^\s;,]* # token that is more than just "q"
        )
      )*                    # zero or more parameters
    )                       # end of media-range
    (?:[ \t]*;[ \t]*q=      # weight is a "q" parameter
      (\d*(?:\.\d+)?)       # qvalue capturing-parentheses
      [^,]*                 # "extension" accept params: who cares?
    )?                      # accept params are optional
    """,
    re.VERBOSE,
)


async def get_accepted_languages(accept_language: str = Header(None)) -> List[str]:
    if accept_language is None:
        return [FALLBACK]

    parsed_languages: List[Tuple[str, float]] = []
    for match in ACCEPT_LANGUAGE_REGEX.finditer(accept_language):
        language = match.group(1).replace("-", "_")
        quality = match.group(2)
        if quality is None:
            quality = 1.0
        else:
            quality = float(quality)
        parsed_languages.append((language, quality))

    sorted_languages = sorted(
        parsed_languages, key=operator.itemgetter(1), reverse=True
    )
    return [language for (language, _) in sorted_languages]


async def get_translations(
    request: Request,
    accepted_languages: List[str] = Depends(get_accepted_languages),
) -> Translations:
    translations = get_preferred_translations(accepted_languages)
    request.scope["translations"] = translations
    return translations


async def get_gettext(translations: Translations = Depends(get_translations)):
    return translations.gettext



from pathlib import Path

from fastapi import FastAPI, Response, status
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles
from starlette.exceptions import HTTPException

from fief.paths import STATIC_DIRECTORY


class StaticFilesFallback(StaticFiles):
    def __init__(self, *args, fallback: Path, **kwargs) -> None:
        self.fallback = fallback
        super().__init__(*args, **kwargs)

    async def get_response(self, *args, **kwargs) -> Response:
        try:
            return await super().get_response(*args, **kwargs)
        except HTTPException as e:
            if e.status_code == status.HTTP_404_NOT_FOUND:
                return FileResponse(self.fallback)
            raise


app = FastAPI()

app.mount(
    "/",
    StaticFilesFallback(
        directory=STATIC_DIRECTORY / "frontend",
        fallback=STATIC_DIRECTORY / "frontend" / "index.html",
        html=False,
        check_dir=False,
    ),
    name="admin_frontend:frontend",
)



import secrets

from fastapi import APIRouter, Depends, HTTPException, status

from fief import schemas
from fief.crypto.jwk import generate_jwk
from fief.dependencies.admin_authentication import is_authenticated_admin
from fief.dependencies.client import get_client_by_id_or_404, get_paginated_clients
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.workspace_managers import get_client_manager, get_tenant_manager
from fief.errors import APIErrorCode
from fief.managers import ClientManager, TenantManager
from fief.models import Client
from fief.schemas.generics import PaginatedResults

router = APIRouter(dependencies=[Depends(is_authenticated_admin)])


@router.get(
    "/", name="clients:list", response_model=PaginatedResults[schemas.client.Client]
)
async def list_clients(
    paginated_clients: PaginatedObjects[Client] = Depends(get_paginated_clients),
) -> PaginatedResults[schemas.client.Client]:
    clients, count = paginated_clients
    return PaginatedResults(
        count=count,
        results=[schemas.client.Client.from_orm(client) for client in clients],
    )


@router.post(
    "/",
    name="clients:create",
    response_model=schemas.client.Client,
    status_code=status.HTTP_201_CREATED,
)
async def create_client(
    client_create: schemas.client.ClientCreate,
    manager: ClientManager = Depends(get_client_manager),
    tenant_manager: TenantManager = Depends(get_tenant_manager),
) -> schemas.client.Client:
    tenant = await tenant_manager.get_by_id(client_create.tenant_id)
    if tenant is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.CLIENT_CREATE_UNKNOWN_TENANT,
        )

    client = Client(**client_create.dict())
    client = await manager.create(client)

    return schemas.client.Client.from_orm(client)


@router.patch("/{id:uuid}", name="clients:update", response_model=schemas.client.Client)
async def update_client(
    client_update: schemas.client.ClientUpdate,
    client: Client = Depends(get_client_by_id_or_404),
    manager: ClientManager = Depends(get_client_manager),
) -> schemas.client.Client:
    client_update_dict = client_update.dict(exclude_unset=True)
    for field, value in client_update_dict.items():
        setattr(client, field, value)

    await manager.update(client)

    return schemas.client.Client.from_orm(client)


@router.post(
    "/{id:uuid}/encryption-key",
    name="clients:encryption_key",
    status_code=status.HTTP_201_CREATED,
)
async def create_encryption_key(
    client: Client = Depends(get_client_by_id_or_404),
    manager: ClientManager = Depends(get_client_manager),
):
    key = generate_jwk(secrets.token_urlsafe(), "enc")
    client.encrypt_jwk = key.export_public()
    await manager.update(client)

    return key.export(as_dict=True)



from datetime import date, datetime
from typing import Any, Dict, Generic, Optional, TypeVar, Union

from fastapi_users import schemas
from pydantic import UUID4, BaseModel, Field, StrictBool, StrictInt, StrictStr
from pydantic.generics import GenericModel

from fief.schemas.generics import Address, CreatedUpdatedAt, Timezone
from fief.schemas.tenant import TenantEmbedded


class UserRead(schemas.BaseUser, CreatedUpdatedAt):
    tenant_id: UUID4
    tenant: TenantEmbedded
    fields: Dict[
        str, Union[Address, Timezone, StrictBool, StrictInt, StrictStr, datetime, date]
    ]

    class Config:
        orm_mode = True


class UserFields(BaseModel):
    def get_value(self, field: str) -> Any:
        value = getattr(self, field)
        if isinstance(value, BaseModel):
            return value.dict()
        return value


UF = TypeVar("UF", bound=UserFields)


class UserCreate(GenericModel, Generic[UF], schemas.BaseUserCreate):
    fields: UF = Field(default_factory=dict, exclude=True)


class UserCreateInternal(UserCreate[UF], Generic[UF]):
    """
    Utility model so that we can hook into the logic of UserManager.create
    and add some attributes before persisting into database.
    """

    tenant_id: UUID4


class UserUpdate(GenericModel, Generic[UF], schemas.BaseUserUpdate):
    fields: Optional[UF] = Field(exclude=True)



import asyncio
import contextlib
from typing import Any, AsyncContextManager, AsyncGenerator, Callable, ClassVar, Dict
from urllib.parse import urlparse

import dramatiq
import jinja2
from dramatiq.brokers.redis import RedisBroker
from pydantic import UUID4
from sqlalchemy import select

from fief.db import AsyncSession
from fief.db.engine import AsyncSession, create_async_session_maker, create_engine
from fief.db.workspace import get_connection
from fief.locale import Translations
from fief.managers import TenantManager, WorkspaceManager
from fief.models import Tenant, User, Workspace
from fief.paths import EMAIL_TEMPLATES_DIRECTORY
from fief.services.email import EmailProvider
from fief.settings import settings

redis_parameters = urlparse(settings.redis_url)
redis_broker = RedisBroker(
    host=redis_parameters.hostname,
    port=redis_parameters.port,
    username=redis_parameters.username,
    password=redis_parameters.password,
    # Heroku Redis with TLS use self-signed certs, so we need to tinker a bit
    ssl=redis_parameters.scheme == "rediss",
    ssl_cert_reqs=None,
)
dramatiq.set_broker(redis_broker)


SendTask = Callable[..., None]


def send_task(task: dramatiq.Actor, *args, **kwargs):
    task.send(*args, **kwargs)


def get_main_session_task() -> AsyncContextManager[AsyncSession]:
    main_engine = create_engine(settings.get_database_url())
    return create_async_session_maker(main_engine)()


@contextlib.asynccontextmanager
async def get_workspace_session_task(
    workspace: Workspace,
) -> AsyncGenerator[AsyncSession, None]:
    engine = create_engine(workspace.get_database_url())
    async with get_connection(engine, workspace.get_schema_name()) as connection:
        async with AsyncSession(bind=connection, expire_on_commit=False) as session:
            yield session


email_provider = settings.get_email_provider()


class TaskError(Exception):
    pass


class TaskBase:
    __name__: ClassVar[str]

    def __init__(
        self,
        get_main_session: Callable[
            ..., AsyncContextManager[AsyncSession]
        ] = get_main_session_task,
        get_workspace_session: Callable[
            ..., AsyncContextManager[AsyncSession]
        ] = get_workspace_session_task,
        email_provider: EmailProvider = email_provider,
    ) -> None:
        self.get_main_session = get_main_session
        self.get_workspace_session = get_workspace_session
        self.email_provider = email_provider

        self.jinja_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(EMAIL_TEMPLATES_DIRECTORY), autoescape=True
        )
        self.jinja_env.add_extension("jinja2.ext.i18n")

    def __call__(self, *args, **kwargs):
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # The default policy doesn't create a loop by default for threads (only for main process)
            # Thus, we create one here and set it for future works.
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        return loop.run_until_complete(self.run(*args, **kwargs))

    async def _get_workspace(self, workspace_id: UUID4) -> Workspace:
        async with self.get_main_session() as session:
            manager = WorkspaceManager(session)
            workspace = await manager.get_by_id(workspace_id)
            if workspace is None:
                raise TaskError()
            return workspace

    async def _get_user(self, user_id: UUID4, workspace: Workspace) -> User:
        async with self.get_workspace_session(workspace) as session:
            statement = select(User).where(User.id == user_id)
            results = await session.execute(statement)
            row = results.first()
            if row is None:
                raise TaskError()
            user = row[0]
            return user

    async def _get_tenant(self, tenant_id: UUID4, workspace: Workspace) -> Tenant:
        async with self.get_workspace_session(workspace) as session:
            manager = TenantManager(session)
            tenant = await manager.get_by_id(tenant_id)
            if tenant is None:
                raise TaskError()
            return tenant

    def _render_email_template(
        self, template: str, translations: Translations, context: Dict[str, Any]
    ) -> str:
        self.jinja_env.install_gettext_translations(translations, newstyle=True)  # type: ignore
        template_object = self.jinja_env.get_template(template)
        return template_object.render(context)



from fastapi import APIRouter, Depends, status

from fief import schemas
from fief.crypto.token import generate_token
from fief.dependencies.admin_api_key import (
    get_api_key_by_id_or_404,
    get_paginated_api_keys,
)
from fief.dependencies.admin_session import get_admin_session_token
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.main_managers import get_admin_api_key_manager
from fief.dependencies.pagination import PaginatedObjects
from fief.managers import AdminAPIKeyManager
from fief.models import AdminAPIKey, Workspace
from fief.schemas.generics import PaginatedResults

router = APIRouter()


@router.get("/", name="api_keys:list", dependencies=[Depends(get_admin_session_token)])
async def list_api_keys(
    paginated_api_keys: PaginatedObjects[AdminAPIKey] = Depends(get_paginated_api_keys),
) -> PaginatedResults[schemas.admin_api_key.AdminAPIKey]:
    api_keys, count = paginated_api_keys
    return PaginatedResults(
        count=count,
        results=[
            schemas.admin_api_key.AdminAPIKey.from_orm(api_key) for api_key in api_keys
        ],
    )


@router.post(
    "/",
    name="api_keys:create",
    dependencies=[Depends(get_admin_session_token)],
    status_code=status.HTTP_201_CREATED,
)
async def create_api_key(
    create_api_key: schemas.admin_api_key.AdminAPIKeyCreate,
    current_workspace: Workspace = Depends(get_current_workspace),
    manager: AdminAPIKeyManager = Depends(get_admin_api_key_manager),
) -> schemas.admin_api_key.AdminAPIKeyCreateResponse:
    token, token_hash = generate_token()
    api_key = AdminAPIKey(
        **create_api_key.dict(), token=token_hash, workspace_id=current_workspace.id
    )
    api_key = await manager.create(api_key)

    api_key_response = schemas.admin_api_key.AdminAPIKeyCreateResponse.from_orm(api_key)
    api_key_response.token = token
    return api_key_response


@router.delete(
    "/{id:uuid}",
    name="api_keys:delete",
    dependencies=[Depends(get_admin_session_token)],
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_api_key(
    api_key: AdminAPIKey = Depends(get_api_key_by_id_or_404),
    manager: AdminAPIKeyManager = Depends(get_admin_api_key_manager),
) -> None:
    await manager.delete(api_key)
    return None



from fastapi import Depends

from fief.db import AsyncSession
from fief.db.main import get_main_async_session
from fief.managers import (
    AdminAPIKeyManager,
    AdminSessionTokenManager,
    WorkspaceManager,
    WorkspaceUserManager,
    get_manager,
)


async def get_workspace_manager(
    session: AsyncSession = Depends(get_main_async_session),
) -> WorkspaceManager:
    return get_manager(WorkspaceManager, session)


async def get_admin_session_token_manager(
    session: AsyncSession = Depends(get_main_async_session),
) -> AdminSessionTokenManager:
    return get_manager(AdminSessionTokenManager, session)


async def get_admin_api_key_manager(
    session: AsyncSession = Depends(get_main_async_session),
) -> AdminAPIKeyManager:
    return get_manager(AdminAPIKeyManager, session)


async def get_workspace_user_manager(
    session: AsyncSession = Depends(get_main_async_session),
) -> WorkspaceUserManager:
    return get_manager(WorkspaceUserManager, session)



import functools
import json
import uuid
from typing import Any, Dict

from fief_client import FiefTokenResponse
from pydantic import UUID4
from sqlalchemy import Column, String, Text

from fief.models.base import MainBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel


class AdminSessionToken(UUIDModel, CreatedUpdatedAt, MainBase):
    __tablename__ = "admin_session_tokens"

    token: str = Column(String(length=255), unique=True, nullable=False)
    raw_tokens: str = Column(Text, nullable=False)
    raw_userinfo: str = Column(Text, nullable=False)

    def __repr__(self) -> str:
        return f"AdminSessionToken(id={self.id})"

    @functools.cached_property
    def tokens(self) -> FiefTokenResponse:
        return json.loads(self.raw_tokens)

    @functools.cached_property
    def userinfo(self) -> Dict[str, Any]:
        return json.loads(self.raw_userinfo)

    @functools.cached_property
    def user_id(self) -> UUID4:
        return uuid.UUID(self.userinfo["sub"])



from typing import Any, Dict

from fastapi import Request


async def get_form_data_dict(request: Request) -> Dict[str, Any]:
    form = await request.form()
    form_data_dict: Dict[str, Any] = {}

    for field in form:
        data_dict = form_data_dict
        path = field.split(".")
        for key in path[:-1]:
            data_dict = data_dict.setdefault(key, {})
        if form[field] != "":
            data_dict[path[-1]] = form[field]

    return form_data_dict



from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import User


class UserManager(BaseManager[User], UUIDManagerMixin[User]):
    model = User

    async def count_all(self) -> int:
        statement = select(User)
        return await self._count(statement)



from typing import Optional

from fastapi import Cookie, Depends

from fief.crypto.token import get_token_hash
from fief.dependencies.workspace_managers import get_session_token_manager
from fief.managers import SessionTokenManager
from fief.models import SessionToken
from fief.settings import settings


async def get_session_token(
    token: Optional[str] = Cookie(None, alias=settings.session_cookie_name),
    manager: SessionTokenManager = Depends(get_session_token_manager),
) -> Optional[SessionToken]:
    if token is not None:
        token_hash = get_token_hash(token)
        return await manager.get_by_token(token_hash)
    return None



"""Initial migration

Revision ID: bc4021719fac
Revises:
Create Date: 2022-04-13 15:36:14.370753

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

import fief

# revision identifiers, used by Alembic.
revision = "bc4021719fac"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    if connection.dialect.name == "postgresql":
        client_type_enum = postgresql.ENUM(
            "public", "confidential", name="clienttype", create_type=False
        )
        client_type_enum.create(connection, checkfirst=True)
    else:
        client_type_enum = sa.Enum("public", "confidential", name="clienttype")

    op.create_table(
        "fief_tenants",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("slug", sa.String(length=255), nullable=False),
        sa.Column("default", sa.Boolean(), nullable=False),
        sa.Column("sign_jwk", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("slug"),
    )
    op.create_index(
        op.f("ix_fief_tenants_created_at"), "fief_tenants", ["created_at"], unique=False
    )
    op.create_index(
        op.f("ix_fief_tenants_updated_at"), "fief_tenants", ["updated_at"], unique=False
    )
    op.create_table(
        "fief_clients",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("first_party", sa.Boolean(), nullable=False),
        sa.Column(
            "client_type",
            client_type_enum,
            server_default="confidential",
            nullable=False,
        ),
        sa.Column("client_id", sa.String(length=255), nullable=False),
        sa.Column("client_secret", sa.String(length=255), nullable=False),
        sa.Column("redirect_uris", sa.JSON(), nullable=False),
        sa.Column("encrypt_jwk", sa.Text(), nullable=True),
        sa.Column("tenant_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(["tenant_id"], ["fief_tenants.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_clients_client_id"), "fief_clients", ["client_id"], unique=False
    )
    op.create_index(
        op.f("ix_fief_clients_client_secret"),
        "fief_clients",
        ["client_secret"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_clients_created_at"), "fief_clients", ["created_at"], unique=False
    )
    op.create_index(
        op.f("ix_fief_clients_updated_at"), "fief_clients", ["updated_at"], unique=False
    )
    op.create_table(
        "fief_users",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("email", sa.String(length=320), nullable=False),
        sa.Column("hashed_password", sa.String(length=255), nullable=False),
        sa.Column("is_active", sa.Boolean(), nullable=False),
        sa.Column("is_superuser", sa.Boolean(), nullable=False),
        sa.Column("is_verified", sa.Boolean(), nullable=False),
        sa.Column("tenant_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(["tenant_id"], ["fief_tenants.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("email", "tenant_id"),
    )
    op.create_index(
        op.f("ix_fief_users_created_at"), "fief_users", ["created_at"], unique=False
    )
    op.create_index(op.f("ix_fief_users_email"), "fief_users", ["email"], unique=False)
    op.create_index(
        op.f("ix_fief_users_updated_at"), "fief_users", ["updated_at"], unique=False
    )
    op.create_table(
        "fief_authorization_codes",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("code", sa.String(length=255), nullable=False),
        sa.Column("c_hash", sa.String(length=255), nullable=False),
        sa.Column("redirect_uri", sa.String(length=2048), nullable=False),
        sa.Column("scope", sa.JSON(), nullable=False),
        sa.Column(
            "authenticated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=False,
        ),
        sa.Column("nonce", sa.String(length=255), nullable=True),
        sa.Column("code_challenge", sa.String(length=255), nullable=True),
        sa.Column("code_challenge_method", sa.String(length=255), nullable=True),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("client_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(["client_id"], ["fief_clients.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_id"], ["fief_users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_authorization_codes_code"),
        "fief_authorization_codes",
        ["code"],
        unique=True,
    )
    op.create_index(
        op.f("ix_fief_authorization_codes_created_at"),
        "fief_authorization_codes",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_authorization_codes_updated_at"),
        "fief_authorization_codes",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_grants",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("scope", sa.JSON(), nullable=False),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("client_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(["client_id"], ["fief_clients.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_id"], ["fief_users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("user_id", "client_id"),
    )
    op.create_index(
        op.f("ix_fief_grants_created_at"), "fief_grants", ["created_at"], unique=False
    )
    op.create_index(
        op.f("ix_fief_grants_updated_at"), "fief_grants", ["updated_at"], unique=False
    )
    op.create_table(
        "fief_login_sessions",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("token", sa.String(length=255), nullable=False),
        sa.Column("response_type", sa.String(length=255), nullable=False),
        sa.Column("response_mode", sa.String(length=255), nullable=False),
        sa.Column("redirect_uri", sa.String(length=2048), nullable=False),
        sa.Column("scope", sa.JSON(), nullable=False),
        sa.Column("prompt", sa.String(length=255), nullable=True),
        sa.Column("state", sa.String(length=2048), nullable=True),
        sa.Column("nonce", sa.String(length=255), nullable=True),
        sa.Column("code_challenge", sa.String(length=255), nullable=True),
        sa.Column("code_challenge_method", sa.String(length=255), nullable=True),
        sa.Column("client_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(["client_id"], ["fief_clients.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_login_sessions_created_at"),
        "fief_login_sessions",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_login_sessions_token"),
        "fief_login_sessions",
        ["token"],
        unique=True,
    )
    op.create_index(
        op.f("ix_fief_login_sessions_updated_at"),
        "fief_login_sessions",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_refresh_tokens",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("token", sa.String(length=255), nullable=False),
        sa.Column(
            "expires_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=False,
        ),
        sa.Column("scope", sa.JSON(), nullable=False),
        sa.Column(
            "authenticated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=False,
        ),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("client_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(["client_id"], ["fief_clients.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["user_id"], ["fief_users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_refresh_tokens_created_at"),
        "fief_refresh_tokens",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_refresh_tokens_expires_at"),
        "fief_refresh_tokens",
        ["expires_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_refresh_tokens_token"),
        "fief_refresh_tokens",
        ["token"],
        unique=True,
    )
    op.create_index(
        op.f("ix_fief_refresh_tokens_updated_at"),
        "fief_refresh_tokens",
        ["updated_at"],
        unique=False,
    )
    op.create_table(
        "fief_session_tokens",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("token", sa.String(length=255), nullable=False),
        sa.Column(
            "expires_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=False,
        ),
        sa.Column("user_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(["user_id"], ["fief_users.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_session_tokens_created_at"),
        "fief_session_tokens",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_session_tokens_expires_at"),
        "fief_session_tokens",
        ["expires_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_session_tokens_token"),
        "fief_session_tokens",
        ["token"],
        unique=True,
    )
    op.create_index(
        op.f("ix_fief_session_tokens_updated_at"),
        "fief_session_tokens",
        ["updated_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_fief_session_tokens_updated_at"), table_name="fief_session_tokens"
    )
    op.drop_index(
        op.f("ix_fief_session_tokens_token"), table_name="fief_session_tokens"
    )
    op.drop_index(
        op.f("ix_fief_session_tokens_expires_at"), table_name="fief_session_tokens"
    )
    op.drop_index(
        op.f("ix_fief_session_tokens_created_at"), table_name="fief_session_tokens"
    )
    op.drop_table("fief_session_tokens")
    op.drop_index(
        op.f("ix_fief_refresh_tokens_updated_at"), table_name="fief_refresh_tokens"
    )
    op.drop_index(
        op.f("ix_fief_refresh_tokens_token"), table_name="fief_refresh_tokens"
    )
    op.drop_index(
        op.f("ix_fief_refresh_tokens_expires_at"), table_name="fief_refresh_tokens"
    )
    op.drop_index(
        op.f("ix_fief_refresh_tokens_created_at"), table_name="fief_refresh_tokens"
    )
    op.drop_table("fief_refresh_tokens")
    op.drop_index(
        op.f("ix_fief_login_sessions_updated_at"), table_name="fief_login_sessions"
    )
    op.drop_index(
        op.f("ix_fief_login_sessions_token"), table_name="fief_login_sessions"
    )
    op.drop_index(
        op.f("ix_fief_login_sessions_created_at"), table_name="fief_login_sessions"
    )
    op.drop_table("fief_login_sessions")
    op.drop_index(op.f("ix_fief_grants_updated_at"), table_name="fief_grants")
    op.drop_index(op.f("ix_fief_grants_created_at"), table_name="fief_grants")
    op.drop_table("fief_grants")
    op.drop_index(
        op.f("ix_fief_authorization_codes_updated_at"),
        table_name="fief_authorization_codes",
    )
    op.drop_index(
        op.f("ix_fief_authorization_codes_created_at"),
        table_name="fief_authorization_codes",
    )
    op.drop_index(
        op.f("ix_fief_authorization_codes_code"), table_name="fief_authorization_codes"
    )
    op.drop_table("fief_authorization_codes")
    op.drop_index(op.f("ix_fief_users_updated_at"), table_name="fief_users")
    op.drop_index(op.f("ix_fief_users_email"), table_name="fief_users")
    op.drop_index(op.f("ix_fief_users_created_at"), table_name="fief_users")
    op.drop_table("fief_users")
    op.drop_index(op.f("ix_fief_clients_updated_at"), table_name="fief_clients")
    op.drop_index(op.f("ix_fief_clients_created_at"), table_name="fief_clients")
    op.drop_index(op.f("ix_fief_clients_client_secret"), table_name="fief_clients")
    op.drop_index(op.f("ix_fief_clients_client_id"), table_name="fief_clients")
    op.drop_table("fief_clients")
    op.drop_index(op.f("ix_fief_tenants_updated_at"), table_name="fief_tenants")
    op.drop_index(op.f("ix_fief_tenants_created_at"), table_name="fief_tenants")
    op.drop_table("fief_tenants")

    try:
        client_type_enum = postgresql.ENUM(
            "public", "confidential", name="clienttype", create_type=False
        )
        client_type_enum.drop(op.get_bind(), checkfirst=True)
    except AttributeError:
        pass
    # ### end Alembic commands ###



from fastapi.templating import Jinja2Templates
from pycountry import countries
from pytz import common_timezones
from starlette.background import BackgroundTasks

from fief.locale import Translations
from fief.paths import TEMPLATES_DIRECTORY


class LocaleJinja2Templates(Jinja2Templates):
    def _create_env(self, directory):
        env = super()._create_env(directory)
        env.add_extension("jinja2.ext.i18n")

        env.globals["countries"] = sorted(countries, key=lambda c: c.name)
        env.globals["timezones"] = sorted(common_timezones)

        return env

    def LocaleTemplateResponse(
        self,
        name: str,
        context: dict,
        translations: Translations,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        background: BackgroundTasks = None,
    ):
        self.env.install_gettext_translations(translations, newstyle=True)  # type: ignore
        return super().TemplateResponse(
            name, context, status_code, headers, media_type, background
        )


templates = LocaleJinja2Templates(directory=TEMPLATES_DIRECTORY)



from typing import Dict, List, Optional, Type, Union, cast

from fastapi import Depends, HTTPException, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.security import OAuth2AuthorizationCodeBearer
from fastapi_users import BaseUserManager, InvalidPasswordException
from fastapi_users.authentication import (
    AuthenticationBackend,
    BearerTransport,
    Strategy,
)
from fastapi_users.exceptions import UserNotExists
from fastapi_users.manager import UUIDIDMixin
from fastapi_users.password import PasswordHelperProtocol
from fastapi_users_db_sqlalchemy import (
    SQLAlchemyBaseOAuthAccountTable,
    SQLAlchemyUserDatabase,
)
from furl import furl
from jwcrypto import jwk
from pydantic import UUID4, ValidationError, create_model
from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import Select

from fief.crypto.access_token import InvalidAccessToken, read_access_token
from fief.crypto.password import password_helper
from fief.db import AsyncSession
from fief.dependencies.current_workspace import (
    get_current_workspace,
    get_current_workspace_session,
)
from fief.dependencies.locale import get_translations
from fief.dependencies.pagination import (
    Ordering,
    PaginatedObjects,
    Pagination,
    get_ordering,
    get_paginated_objects,
    get_pagination,
)
from fief.dependencies.tasks import get_send_task
from fief.dependencies.tenant import (
    get_current_tenant,
    get_tenant_from_create_user_internal,
)
from fief.dependencies.user_field import (
    get_admin_user_create_internal_model,
    get_admin_user_update_model,
    get_user_update_model,
)
from fief.dependencies.workspace_managers import get_user_manager as get_user_db_manager
from fief.locale import Translations
from fief.managers import UserManager as UserDBManager
from fief.models import Tenant, User, UserField, UserFieldValue, Workspace
from fief.schemas.user import UF, UserCreate, UserCreateInternal, UserUpdate
from fief.settings import settings
from fief.tasks import SendTask, on_after_forgot_password, on_after_register


class UserManager(UUIDIDMixin, BaseUserManager[User, UUID4]):
    reset_password_token_secret = settings.secret.get_secret_value()
    verification_token_secret = settings.secret.get_secret_value()

    def __init__(
        self,
        user_db: SQLAlchemyUserDatabase[User, UUID4],
        password_helper: PasswordHelperProtocol,
        workspace: Workspace,
        tenant: Tenant,
        translations: Translations,
        send_task: SendTask,
    ):
        super().__init__(user_db, password_helper)
        self.workspace = workspace
        self.tenant = tenant
        self.translations = translations
        self.send_task = send_task

    async def validate_password(  # type: ignore
        self, password: str, user: Union[UserCreate, User]
    ) -> None:
        if len(password) < 8:
            raise InvalidPasswordException(
                reason=self.translations.gettext(
                    "The password should be at least 8 characters."
                )
            )

    async def create_with_fields(
        self,
        user_create: UserCreate[UF],
        *,
        user_fields: List[UserField],
        safe: bool = False,
        request: Optional[Request] = None,
    ) -> User:
        user = await self.create(user_create, safe, request)

        for user_field in user_fields:
            user_field_value = UserFieldValue(user_field=user_field)
            try:
                value = user_create.fields.get_value(user_field.slug)
                if value is not None:
                    user_field_value.value = value
                    user.user_field_values.append(user_field_value)
            except AttributeError:
                default = user_field.get_default()
                if default is not None:
                    user_field_value.value = default
                    user.user_field_values.append(user_field_value)

        return await self.user_db.update(user, {})

    async def update_with_fields(
        self,
        user_update: UserUpdate[UF],
        user: User,
        *,
        user_fields: List[UserField],
        safe: bool = False,
        request: Optional[Request] = None,
    ) -> User:
        user = await self.update(user_update, user, safe, request)

        if user_update.fields is not None:
            for user_field in user_fields:
                existing_user_field_value = user.get_user_field_value(user_field)
                # Update existing value
                if existing_user_field_value is not None:
                    try:
                        value = user_update.fields.get_value(user_field.slug)
                        if value is not None:
                            existing_user_field_value.value = value
                    except AttributeError:
                        pass
                # Create new value
                else:
                    user_field_value = UserFieldValue(user_field=user_field)
                    try:
                        value = user_update.fields.get_value(user_field.slug)
                        if value is not None:
                            user_field_value.value = value
                            user.user_field_values.append(user_field_value)
                    except AttributeError:
                        default = user_field.get_default()
                        if default is not None:
                            user_field_value.value = default
                            user.user_field_values.append(user_field_value)

            user = await self.user_db.update(user, {})

        return user

    async def on_after_register(self, user: User, request: Optional[Request] = None):
        self.send_task(on_after_register, str(user.id), str(self.workspace.id))

    async def on_after_forgot_password(
        self, user: User, token: str, request: Optional[Request] = None
    ):
        reset_url = furl(self.tenant.url_for(cast(Request, request), "reset:reset.get"))
        reset_url.add(query_params={"token": token})
        self.send_task(
            on_after_forgot_password,
            str(user.id),
            str(self.workspace.id),
            reset_url.url,
        )

    async def on_after_request_verify(
        self, user: User, token: str, request: Optional[Request] = None
    ):
        print(f"Verification requested for user {user.id}. Verification token: {token}")


class SQLAlchemyUserTenantDatabase(SQLAlchemyUserDatabase[User, UUID4]):
    def __init__(
        self,
        session: AsyncSession,
        tenant: Tenant,
        user_table: Type[User],
        oauth_account_table: Optional[Type[SQLAlchemyBaseOAuthAccountTable]] = None,
    ):
        super().__init__(session, user_table, oauth_account_table=oauth_account_table)
        self.tenant = tenant

    async def _get_user(self, statement: Select) -> Optional[User]:
        statement = statement.where(User.tenant_id == self.tenant.id)
        return await super()._get_user(statement)


class JWTAccessTokenStrategy(Strategy[User, UUID4]):
    def __init__(self, key: jwk.JWK):
        self.key = key

    async def read_token(
        self, token: Optional[str], user_manager: BaseUserManager[User, UUID4]
    ) -> Optional[User]:
        if token is None:
            return None

        try:
            user_id = read_access_token(self.key, token)
            user = await user_manager.get(user_id)
        except InvalidAccessToken:
            return None
        except UserNotExists:
            return None
        else:
            return user


async def get_jwt_access_token_strategy(
    tenant: Tenant = Depends(get_current_tenant),
) -> JWTAccessTokenStrategy:
    return JWTAccessTokenStrategy(tenant.get_sign_jwk())


async def get_user_db(
    session: AsyncSession = Depends(get_current_workspace_session),
    tenant: Tenant = Depends(get_current_tenant),
) -> SQLAlchemyUserDatabase[User, UUID4]:
    return SQLAlchemyUserTenantDatabase(session, tenant, User)


async def get_user_manager(
    user_db: SQLAlchemyUserDatabase[User, UUID4] = Depends(get_user_db),
    tenant: Tenant = Depends(get_current_tenant),
    workspace: Workspace = Depends(get_current_workspace),
    translations: Translations = Depends(get_translations),
    send_task: SendTask = Depends(get_send_task),
):
    return UserManager(
        user_db, password_helper, workspace, tenant, translations, send_task
    )


async def get_user_db_from_create_user_internal(
    session: AsyncSession = Depends(get_current_workspace_session),
    tenant: Tenant = Depends(get_tenant_from_create_user_internal),
) -> SQLAlchemyUserDatabase[User, UUID4]:
    return SQLAlchemyUserTenantDatabase(session, tenant, User)


async def get_user_manager_from_create_user_internal(
    user_db: SQLAlchemyUserDatabase[User, UUID4] = Depends(
        get_user_db_from_create_user_internal
    ),
    tenant: Tenant = Depends(get_tenant_from_create_user_internal),
    workspace: Workspace = Depends(get_current_workspace),
    translations: Translations = Depends(get_translations),
    send_task: SendTask = Depends(get_send_task),
):
    return UserManager(
        user_db, password_helper, workspace, tenant, translations, send_task
    )


class AuthorizationCodeBearerTransport(BearerTransport):
    scheme: OAuth2AuthorizationCodeBearer  # type: ignore

    def __init__(
        self,
        authorizationUrl: str,
        tokenUrl: str,
        refreshUrl: str,
        scopes: Dict[str, str],
    ):
        self.scheme = OAuth2AuthorizationCodeBearer(
            authorizationUrl, tokenUrl, refreshUrl, scopes=scopes
        )


authentication_backend = AuthenticationBackend[User, UUID4](
    "jwt_access_token",
    AuthorizationCodeBearerTransport(
        "/authorize", "/api/token", "/api/token", {"openid": "openid"}
    ),
    get_jwt_access_token_strategy,
)


async def get_paginated_users(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(get_ordering),
    manager: UserDBManager = Depends(get_user_db_manager),
) -> PaginatedObjects[User]:
    statement = select(User).options(joinedload(User.tenant))
    return await get_paginated_objects(statement, pagination, ordering, manager)


async def get_user_by_id_or_404(
    id: UUID4,
    manager: UserDBManager = Depends(get_user_db_manager),
) -> User:
    user = await manager.get_by_id(id, (joinedload(User.tenant),))

    if user is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return user


async def get_user_db_from_user(
    user: User = Depends(get_user_by_id_or_404),
    session: AsyncSession = Depends(get_current_workspace_session),
) -> SQLAlchemyUserDatabase[User, UUID4]:
    return SQLAlchemyUserTenantDatabase(session, user.tenant, User)


async def get_user_manager_from_user(
    user: User = Depends(get_user_by_id_or_404),
    user_db: SQLAlchemyUserDatabase[User, UUID4] = Depends(get_user_db_from_user),
    workspace: Workspace = Depends(get_current_workspace),
    translations: Translations = Depends(get_translations),
    send_task: SendTask = Depends(get_send_task),
):
    return UserManager(
        user_db, password_helper, workspace, user.tenant, translations, send_task
    )


async def get_user_create_internal(
    request: Request,
    user_create_internal_model: Type[UserCreateInternal[UF]] = Depends(
        get_admin_user_create_internal_model
    ),
) -> UserCreateInternal[UF]:
    body_model = create_model(
        "UserCreateInternalBody",
        body=(user_create_internal_model, ...),
    )
    try:
        validated_user_create_internal = body_model(body=await request.json())
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e
    else:
        return validated_user_create_internal.body  # type: ignore


async def get_user_update(
    request: Request,
    user_update_model: Type[UserUpdate[UF]] = Depends(get_user_update_model),
) -> UserUpdate[UF]:
    body_model = create_model(
        "UserUpdateBody",
        body=(user_update_model, ...),
    )
    try:
        validated_user_update = body_model(body=await request.json())
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e
    else:
        return validated_user_update.body  # type: ignore


async def get_admin_user_update(
    request: Request,
    user_update_model: Type[UserUpdate[UF]] = Depends(get_admin_user_update_model),
) -> UserUpdate[UF]:
    body_model = create_model(
        "UserUpdateBody",
        body=(user_update_model, ...),
    )
    try:
        validated_user_update = body_model(body=await request.json())
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e
    else:
        return validated_user_update.body  # type: ignore



import functools
import http.cookies
import secrets
from typing import Optional

from fastapi import Cookie, Form, HTTPException, Request, status
from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from fief.errors import APIErrorCode
from fief.settings import settings

CSRF_ATTRIBUTE_NAME = "csrftoken"
SAFE_METHODS = {"GET", "HEAD", "OPTIONS", "TRACE"}


async def check_csrf(
    request: Request,
    challenge_csrf_token: Optional[str] = Cookie(
        None, alias=settings.csrf_cookie_name, include_in_schema=False
    ),
    submitted_csrf_token: Optional[str] = Form(
        None, alias=CSRF_ATTRIBUTE_NAME, include_in_schema=False
    ),
):
    if request.method not in SAFE_METHODS:
        if (
            challenge_csrf_token is None
            or submitted_csrf_token is None
            or not secrets.compare_digest(challenge_csrf_token, submitted_csrf_token)
        ):
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=APIErrorCode.CSRF_CHECK_FAILED,
            )

    request.scope[CSRF_ATTRIBUTE_NAME] = secrets.token_urlsafe()


class CSRFCookieSetterMiddleware:
    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":  # pragma: no cover
            await self.app(scope, receive, send)
            return

        send = functools.partial(self.send, send=send, scope=scope)
        await self.app(scope, receive, send)

    async def send(self, message: Message, send: Send, scope: Scope) -> None:
        if message["type"] != "http.response.start":
            await send(message)
            return

        if csrftoken := scope.get(CSRF_ATTRIBUTE_NAME):
            message.setdefault("headers", [])
            headers = MutableHeaders(scope=message)

            cookie: http.cookies.BaseCookie = http.cookies.SimpleCookie()
            cookie_name = settings.csrf_cookie_name
            cookie[cookie_name] = csrftoken
            cookie[cookie_name]["secure"] = settings.csrf_cookie_secure
            cookie[cookie_name]["httponly"] = True
            headers.append("set-cookie", cookie.output(header="").strip())

        await send(message)



from fastapi import Depends, Request
from pydantic import ValidationError

from fief.dependencies.locale import get_translations
from fief.dependencies.tenant import get_current_tenant
from fief.exceptions import FormValidationError
from fief.locale import Translations
from fief.models import Tenant
from fief.schemas.reset import ForgotPasswordRequest, ResetPasswordRequest


async def get_forgot_password_request(
    request: Request,
    tenant: Tenant = Depends(get_current_tenant),
    translations: Translations = Depends(get_translations),
) -> ForgotPasswordRequest:
    form = await request.form()
    try:
        return ForgotPasswordRequest(**form)
    except ValidationError as e:
        raise FormValidationError(
            "forgot_password.html",
            {"form_data": form, "tenant": tenant},
            translations,
            e.raw_errors,
            ForgotPasswordRequest,
        ) from e


async def get_reset_password_request(
    request: Request,
    tenant: Tenant = Depends(get_current_tenant),
    translations: Translations = Depends(get_translations),
) -> ResetPasswordRequest:
    form = await request.form()
    try:
        return ResetPasswordRequest(**form)
    except ValidationError as e:
        raise FormValidationError(
            "reset_password.html",
            {"form_data": form, "tenant": tenant},
            translations,
            e.raw_errors,
            ResetPasswordRequest,
        ) from e



import gettext
from typing import Dict, List

from babel import core

from fief.paths import LOCALE_DIRECTORY

Translations = gettext.NullTranslations

LOCALES = ["en_US", "fr_FR"]
FALLBACK = "en_US"
TERRITORY_FALLBACKS = {"en_US": "en", "fr_FR": "fr"}
SUPPORTED_LOCALES = LOCALES + list(TERRITORY_FALLBACKS.values())

TRANSLATIONS: Dict[str, Translations] = {}
for locale in LOCALES:
    translation = gettext.translation(
        domain="messages",
        localedir=LOCALE_DIRECTORY,
        languages=[locale],
        fallback=True,
    )
    TRANSLATIONS[locale] = translation
    try:
        TRANSLATIONS[TERRITORY_FALLBACKS[locale]] = translation
    except KeyError:
        pass


def get_preferred_locale(
    preferred: List[str],
    *,
    supported: List[str] = SUPPORTED_LOCALES,
    fallback: str = FALLBACK
) -> str:
    locale = core.negotiate_locale(preferred, supported)
    if locale is None:
        return fallback
    return locale


def get_preferred_translations(
    preferred: List[str],
    *,
    supported: List[str] = SUPPORTED_LOCALES,
    fallback: str = FALLBACK
) -> Translations:
    locale = get_preferred_locale(preferred, supported=supported, fallback=fallback)
    return TRANSLATIONS[locale]



from typing import AsyncGenerator

from fief.db.engine import AsyncSession, create_async_session_maker, create_engine
from fief.settings import settings

main_engine = create_engine(settings.get_database_url())
main_async_session_maker = create_async_session_maker(main_engine)


async def get_main_async_session() -> AsyncGenerator[AsyncSession, None]:
    async with main_async_session_maker() as session:
        yield session


__all__ = [
    "main_async_session_maker",
    "get_main_async_session",
]



from dataclasses import dataclass
from typing import List, Optional

from fief.db.main import main_async_session_maker
from fief.db.workspace import get_workspace_session
from fief.managers import UserManager, WorkspaceManager
from fief.models import Workspace


@dataclass
class WorkspaceStats:
    workspace: Workspace
    reachable: bool
    external_db: bool
    nb_users: Optional[int] = None


class Workspaces:
    def __init__(self, workspace_manager: WorkspaceManager) -> None:
        self.workspace_manager = workspace_manager

    async def get_stats(self) -> List[WorkspaceStats]:
        stats: List[WorkspaceStats] = []

        workspaces = await self.workspace_manager.all()
        for workspace in workspaces:
            try:
                async with get_workspace_session(workspace) as session:
                    user_manager = UserManager(session)
                    nb_users = await user_manager.count_all()
                    stats.append(
                        WorkspaceStats(
                            workspace=workspace,
                            reachable=True,
                            external_db=workspace.database_type != None,
                            nb_users=nb_users,
                        )
                    )
            except ConnectionError:
                stats.append(
                    WorkspaceStats(
                        workspace=workspace,
                        external_db=workspace.database_type != None,
                        reachable=False,
                    )
                )

        return stats


async def get_workspaces_stats() -> List[WorkspaceStats]:
    async with main_async_session_maker() as session:
        workspace_manager = WorkspaceManager(session)
        workspaces = Workspaces(workspace_manager)
        return await workspaces.get_stats()



import logging

from uvicorn.logging import DefaultFormatter

from fief.settings import settings

formatter = DefaultFormatter(fmt="%(levelprefix)s %(message)s")
handler = logging.StreamHandler()
handler.setFormatter(formatter)

logger = logging.getLogger("fief")
logger.setLevel(settings.log_level)
logger.addHandler(handler)



from datetime import date, datetime
from enum import Enum
from typing import Any, Generic, List, Mapping, Optional, Tuple, Type, TypeVar, Union

from pydantic import BaseModel, StrictBool, StrictInt, StrictStr, constr
from pydantic.generics import GenericModel

from fief.models import UserField as UserFieldModel
from fief.models import UserFieldType
from fief.schemas.generics import (
    Address,
    CreatedUpdatedAt,
    PhoneNumber,
    Timezone,
    UUIDSchema,
)

USER_FIELD_TYPE_MAP: Mapping[UserFieldType, Type[Any]] = {
    UserFieldType.STRING: constr(min_length=1),
    UserFieldType.INTEGER: int,
    UserFieldType.BOOLEAN: bool,
    UserFieldType.DATE: date,
    UserFieldType.DATETIME: datetime,
    UserFieldType.CHOICE: str,
    UserFieldType.PHONE_NUMBER: PhoneNumber,
    UserFieldType.ADDRESS: Address,
    UserFieldType.TIMEZONE: Timezone,
}

USER_FIELD_CAN_HAVE_DEFAULT: Mapping[UserFieldType, bool] = {
    UserFieldType.STRING: True,
    UserFieldType.INTEGER: True,
    UserFieldType.BOOLEAN: True,
    UserFieldType.DATE: False,
    UserFieldType.DATETIME: False,
    UserFieldType.CHOICE: True,
    UserFieldType.PHONE_NUMBER: False,
    UserFieldType.ADDRESS: False,
    UserFieldType.TIMEZONE: True,
}

D = TypeVar("D", Timezone, bool, int, str)
UFT = TypeVar("UFT", bound=UserFieldType)


def get_user_field_pydantic_type(field: UserFieldModel) -> Type[Any]:
    if field.type == UserFieldType.CHOICE:
        choices = (
            field.configuration["choices"] if field.configuration["choices"] else []
        )
        return Enum(  # type: ignore
            f"{field.slug.capitalize()}Enum",
            [(value, value) for (value, _) in choices],
            type=str,
        )
    return USER_FIELD_TYPE_MAP[field.type]


class UserFieldConfigurationBase(BaseModel):
    at_registration: bool
    required: bool
    at_update: bool


class UserFieldConfiguration(UserFieldConfigurationBase):
    choices: Optional[List[Tuple[str, str]]]
    default: Optional[Union[Timezone, StrictBool, StrictInt, StrictStr]]


class UserFieldConfigurationDefault(
    GenericModel, Generic[D], UserFieldConfigurationBase
):
    default: Optional[D]


class UserFieldConfigurationChoice(UserFieldConfigurationDefault[str]):
    choices: Optional[List[Tuple[str, str]]]


class UserFieldCreate(BaseModel):
    name: str
    slug: str
    type: UserFieldType
    configuration: UserFieldConfiguration


class UserFieldUpdate(BaseModel):
    name: Optional[str]
    slug: Optional[str]
    configuration: Optional[UserFieldConfiguration]


class BaseUserField(UUIDSchema, CreatedUpdatedAt):
    name: str
    slug: str
    type: UserFieldType
    configuration: UserFieldConfiguration


class UserField(BaseUserField):
    pass



from fastapi import Depends

from fief.dependencies.workspace_managers import (
    get_authorization_code_manager,
    get_grant_manager,
    get_login_session_manager,
    get_session_token_manager,
)
from fief.managers import (
    AuthorizationCodeManager,
    GrantManager,
    LoginSessionManager,
    SessionTokenManager,
)
from fief.services.authentication_flow import AuthenticationFlow


async def get_authentication_flow(
    authorization_code_manager: AuthorizationCodeManager = Depends(
        get_authorization_code_manager
    ),
    login_session_manager: LoginSessionManager = Depends(get_login_session_manager),
    session_token_manager: SessionTokenManager = Depends(get_session_token_manager),
    grant_manager: GrantManager = Depends(get_grant_manager),
) -> AuthenticationFlow:
    return AuthenticationFlow(
        authorization_code_manager,
        login_session_manager,
        session_token_manager,
        grant_manager,
    )



from typing import Optional

from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import AdminSessionToken


class AdminSessionTokenManager(
    BaseManager[AdminSessionToken], UUIDManagerMixin[AdminSessionToken]
):
    model = AdminSessionToken

    async def get_by_token(self, token: str) -> Optional[AdminSessionToken]:
        statement = select(AdminSessionToken).where(AdminSessionToken.token == token)
        return await self.get_one_or_none(statement)



from enum import Enum
from pathlib import Path
from typing import Any, Dict, Optional
from urllib.parse import urlparse

from pydantic import (
    BaseSettings,
    DirectoryPath,
    Field,
    SecretStr,
    root_validator,
    validator,
)
from sqlalchemy import engine

from fief.crypto.encryption import is_valid_key
from fief.db.types import DatabaseType, create_database_url
from fief.services.email import EMAIL_PROVIDERS, AvailableEmailProvider, EmailProvider


class Environment(str, Enum):
    DEVELOPMENT = "development"
    STAGING = "staging"
    PRODUCTION = "production"


class InvalidEncryptionKeyError(ValueError):
    pass


class Settings(BaseSettings):
    environment: Environment = Environment.DEVELOPMENT
    log_level: str = "INFO"
    unit_tests: bool = False
    sentry_dsn_server: Optional[str] = None
    sentry_dsn_worker: Optional[str] = None
    root_domain: str = "localhost:8000"
    allow_origin_regex: str = "http://.*localhost:[0-9]+"
    port: int = 8000

    secret: SecretStr
    encryption_key: bytes

    generated_jwk_size: int = 4096

    database_type: DatabaseType = DatabaseType.SQLITE
    database_url: Optional[str] = None
    database_host: Optional[str] = None
    database_port: Optional[int] = None
    database_username: Optional[str] = None
    database_password: Optional[str] = None
    database_name: Optional[str] = "fief.db"
    database_location: DirectoryPath = Path.cwd()
    database_pool_recycle_seconds: int = 600

    redis_url: str = "redis://localhost:6379"

    email_provider: AvailableEmailProvider = AvailableEmailProvider.NULL
    email_provider_params: Dict[str, Any] = Field(default_factory=dict)

    workspace_table_prefix: str = "fief_"

    csrf_cookie_name: str = "fief_csrftoken"
    csrf_cookie_secure: bool = True

    login_session_cookie_name: str = "fief_login_session"
    login_session_cookie_domain: str = ""
    login_session_cookie_secure: bool = True

    session_cookie_name: str = "fief_session"
    session_cookie_domain: str = ""
    session_cookie_secure: bool = True
    session_lifetime_seconds: int = 86400 * 30

    authorization_code_lifetime_seconds: int = 600
    access_id_token_lifetime_seconds: int = 3600
    refresh_token_lifetime_seconds: int = 3600 * 24 * 30

    fief_domain: str = "localhost:8000"
    fief_client_id: str
    fief_client_secret: str
    fief_encryption_key: Optional[str] = None

    fief_admin_session_cookie_name: str = "fief_admin_session"
    fief_admin_session_cookie_domain: str = ""
    fief_admin_session_cookie_secure: bool = True

    fief_documentation_url: str = "https://docs.fief.dev"

    class Config:
        env_file = ".env"

    @root_validator(pre=True)
    def parse_database_url(cls, values):
        database_url = values.get("database_url")
        if database_url is not None:
            parsed_database_url = urlparse(database_url)
            values["database_host"] = parsed_database_url.hostname
            values["database_port"] = parsed_database_url.port
            values["database_username"] = parsed_database_url.username
            values["database_password"] = parsed_database_url.password
            values["database_name"] = parsed_database_url.path[1:]
        return values

    @validator("encryption_key", pre=True)
    def validate_encryption_key(cls, value: Optional[str]) -> Optional[bytes]:
        if value is None:
            return value

        key = value.encode("utf-8")
        if not is_valid_key(key):
            raise InvalidEncryptionKeyError()

        return key

    @validator("database_port", pre=True)
    def validate_empty_port(cls, value: Optional[str]) -> Optional[str]:
        if value is None or value == "":
            return None
        return value

    def get_database_url(
        self, asyncio: bool = True, schema: Optional[str] = None
    ) -> engine.URL:
        """
        Returns a proper database URL for async or not-async context.

        Some tools like Alembic still require a sync connection.
        """
        return create_database_url(
            self.database_type,
            asyncio=asyncio,
            username=self.database_username,
            password=self.database_password,
            host=self.database_host,
            port=self.database_port,
            database=self.database_name,
            path=settings.database_location,
            schema=schema,
        )

    def get_email_provider(self) -> EmailProvider:
        provider_class = EMAIL_PROVIDERS[self.email_provider]
        return provider_class(**self.email_provider_params)


settings = Settings()



from datetime import datetime

from pydantic import UUID4
from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.orm import relationship

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, TIMESTAMPAware, UUIDModel
from fief.models.user import User


class SessionToken(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "session_tokens"

    token: str = Column(
        String(length=255),
        nullable=False,
        index=True,
        unique=True,
    )
    expires_at: datetime = Column(
        TIMESTAMPAware(timezone=True), nullable=False, index=True
    )

    user_id: UUID4 = Column(GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    user: User = relationship("User", lazy="joined")



import asyncio
import secrets

import typer
import uvicorn
from alembic import command
from alembic.config import Config
from dramatiq import cli as dramatiq_cli
from pydantic import ValidationError
from rich.console import Console
from rich.table import Table
from sqlalchemy import create_engine, select
from sqlalchemy.orm import sessionmaker

from fief import __version__
from fief.crypto.encryption import generate_key
from fief.paths import ALEMBIC_CONFIG_FILE
from fief.services.workspace_db import (
    WorkspaceDatabase,
    WorkspaceDatabaseConnectionError,
)


def get_settings():
    from fief.settings import settings

    return settings


workspaces = typer.Typer(help="Commands to manage workspaces.")


@workspaces.command()
def list():
    """List all workspaces and their number of users."""
    from fief.services.workspaces import get_workspaces_stats

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    stats = loop.run_until_complete(get_workspaces_stats())

    table = Table()
    table.add_column("Name")
    table.add_column("Reachable")
    table.add_column("Database")
    table.add_column("Database type")
    table.add_column("Migration revision")
    table.add_column("Nb. users")
    for stat in stats:
        table.add_row(
            stat.workspace.name,
            "✅" if stat.reachable else "❌",
            "External" if stat.external_db else "Main",
            stat.workspace.database_type if stat.external_db else None,
            stat.workspace.alembic_revision,
            str(stat.nb_users) if stat.reachable else None,
            style="red" if not stat.reachable else None,
        )
    console = Console()
    console.print(table)


@workspaces.command("migrate")
def migrate_workspaces():
    """Apply database migrations to each workspace database."""
    from fief.models import Workspace

    settings = get_settings()

    engine = create_engine(settings.get_database_url(False))
    Session = sessionmaker(engine)
    with Session() as session:
        workspace_db = WorkspaceDatabase()
        workspaces = select(Workspace)
        for [workspace] in session.execute(workspaces):
            assert isinstance(workspace, Workspace)
            typer.secho(f"Migrating {workspace.name}... ", bold=True, nl=False)
            try:
                alembic_revision = workspace_db.migrate(
                    workspace.get_database_url(False), workspace.get_schema_name()
                )
                workspace.alembic_revision = alembic_revision
                session.add(workspace)
                session.commit()
                typer.secho(f"Done!")
            except WorkspaceDatabaseConnectionError:
                typer.secho(f"Failed!", fg="red", err=True)


@workspaces.command("create-main")
def create_main_workspace():
    """Create a main Fief workspace following the environment settings."""
    from fief.services.main_workspace import (
        MainWorkspaceAlreadyExists,
        create_main_fief_workspace,
    )

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        loop.run_until_complete(create_main_fief_workspace())
        typer.echo("Main Fief workspace created")
    except MainWorkspaceAlreadyExists as e:
        typer.echo("Main Fief workspace already exists")
        raise typer.Exit(code=1) from e


@workspaces.command("create-main-user")
def create_main_user(
    user_email: str = typer.Option(..., help="The admin user email"),
    user_password: str = typer.Option(
        ...,
        prompt=True,
        confirmation_prompt=True,
        hide_input=True,
        help="The admin user password",
    ),
):
    """Create a main Fief workspace user."""
    from fief.services.main_workspace import (
        CreateMainFiefUserError,
        create_main_fief_user,
    )

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        loop.run_until_complete(create_main_fief_user(user_email, user_password))
        typer.echo("Main Fief user created")
    except CreateMainFiefUserError as e:
        typer.echo("An error occured")
        raise typer.Exit(code=1) from e


app = typer.Typer(help="Commands to manage your Fief instance.")
app.add_typer(workspaces, name="workspaces")


@app.command()
def quickstart(
    docker: bool = typer.Option(
        False,
        help="Show the Docker command to run the Fief server with required environment variables.",
    ),
    port: int = typer.Option(
        8000, help="Port on which you want to expose the Fief server."
    ),
    host: str = typer.Option(
        "localhost", help="Host on which you want to expose the Fief server."
    ),
    ssl: bool = typer.Option(
        False,
        help="Whether the Fief server will be served over SSL. For local development, it'll likely be false.",
    ),
):
    """Generate secrets and environment variables to help users getting started quickly."""
    typer.secho(
        "⚠️  Be sure to save the generated secrets somewhere safe for subsequent runs. Otherwise, you may lose access to the data.",
        bold=True,
        fg="red",
        err=True,
    )

    environment_variables = {
        "SECRET": secrets.token_urlsafe(64),
        "FIEF_CLIENT_ID": secrets.token_urlsafe(),
        "FIEF_CLIENT_SECRET": secrets.token_urlsafe(),
        "ENCRYPTION_KEY": generate_key().decode("utf-8"),
        "PORT": port,
        "ROOT_DOMAIN": f"{host}:{port}",
        "FIEF_DOMAIN": f"{host}:{port}",
    }
    if not ssl:
        environment_variables.update(
            {
                "CSRF_COOKIE_SECURE": False,
                "LOGIN_SESSION_COOKIE_SECURE": False,
                "SESSION_COOKIE_SECURE": False,
                "FIEF_ADMIN_SESSION_COOKIE_SECURE": False,
            }
        )

    if docker:
        parts = [
            "docker run",
            "--name fief-server",
            f"-p {port}:{port}",
            "-d",
            *[
                f'-e "{name}={value}"'
                for (name, value) in environment_variables.items()
            ],
            "ghcr.io/fief-dev/fief:latest",
        ]
        typer.echo(" \\\n  ".join(parts))
    else:
        for (name, value) in environment_variables.items():
            typer.echo(f"{typer.style(name, bold=True)}: {value}")


@app.command()
def info():
    """Show current Fief version and settings."""
    try:
        settings = get_settings()

        typer.secho(f"Fief version: {__version__}", bold=True)
        typer.secho(f"Settings", bold=True)
        for key, value in settings.dict().items():
            typer.echo(f"{key}: {value}")
    except ValidationError as e:
        typer.secho(
            "❌ Some environment variables are missing or invalid.", bold=True, fg="red"
        )
        typer.secho(
            "Run 'fief quickstart' to help you generate the secrets.", bold=True
        )
        for error in e.errors():
            typer.secho(error)
        raise typer.Exit(code=1) from e


@app.command("migrate")
def migrate_main():
    """Apply database migrations to the main database."""
    settings = get_settings()

    engine = create_engine(settings.get_database_url(False))
    with engine.begin() as connection:
        alembic_config = Config(ALEMBIC_CONFIG_FILE, ini_section="main")
        alembic_config.attributes["connection"] = connection
        command.upgrade(alembic_config, "head")


@app.command("run-server")
def run_server(
    host: str = "0.0.0.0",
    port: int = 8000,
    migrate: bool = typer.Option(
        True,
        help="Run the migrations on main and workspaces databases before starting.",
    ),
):
    """Run the Fief backend server."""
    if migrate:
        migrate_main()
        migrate_workspaces()
    uvicorn.run("fief.app:app", host=host, port=port)


@app.command(
    "run-worker",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
    add_help_option=False,
)
def run_worker(ctx: typer.Context):
    """
    Run the Fief worker.

    Just forwards the options to the Dramatiq CLI.
    """
    parser = dramatiq_cli.make_argument_parser()
    args = parser.parse_args(ctx.args + ["fief.worker"])
    dramatiq_cli.main(args)


if __name__ == "__main__":
    app()



from typing import TYPE_CHECKING, List, Optional, Union

from sqlalchemy import Column, Enum, String, Text, engine, event
from sqlalchemy.orm import relationship

from fief.crypto.encryption import decrypt, encrypt
from fief.db.types import DatabaseType, create_database_url
from fief.models.base import MainBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel
from fief.settings import settings

if TYPE_CHECKING:
    from fief.models.workspace_user import WorkspaceUser


class Workspace(UUIDModel, CreatedUpdatedAt, MainBase):
    __tablename__ = "workspaces"

    name: str = Column(String(length=255), nullable=False)
    domain: str = Column(String(length=255), nullable=False)

    database_type: Optional[DatabaseType] = Column(Enum(DatabaseType), nullable=True)
    database_host: Optional[str] = Column(Text, nullable=True)
    database_port: Optional[str] = Column(Text, nullable=True)
    database_username: Optional[str] = Column(Text, nullable=True)
    database_password: Optional[str] = Column(Text, nullable=True)
    database_name: Optional[str] = Column(Text, nullable=True)

    alembic_revision: Optional[str] = Column(
        String(length=255), nullable=True, index=True
    )

    workspace_users: List["WorkspaceUser"] = relationship(
        "WorkspaceUser", back_populates="workspace", cascade="all, delete"
    )

    def __repr__(self) -> str:
        return f"Workspace(id={self.id}, name={self.name}, domain={self.domain})"

    def get_database_url(self, asyncio=True) -> engine.URL:
        """
        Return the database URL for this workspace.

        If it's not specified on the model, the instance database URL is returned.
        """
        if self.database_type is None:
            url = settings.get_database_url(asyncio, schema=self.get_schema_name())
        else:
            url = create_database_url(
                self.database_type,
                asyncio=asyncio,
                username=self._decrypt_database_setting("database_username"),
                password=self._decrypt_database_setting("database_password"),
                host=self._decrypt_database_setting("database_host"),
                port=self._decrypt_database_port(),
                database=self._decrypt_database_setting("database_name"),
                path=settings.database_location,
                schema=self.get_schema_name(),
            )

        return url

    def get_schema_name(self) -> str:
        """
        Return the SQL schema name where the data is stored.
        """
        return str(self.id)

    def _decrypt_database_setting(self, attribute: str) -> Optional[str]:
        value = getattr(self, attribute)
        if value is None:
            return value
        return decrypt(value, settings.encryption_key)

    def _decrypt_database_port(self) -> Optional[int]:
        value = self._decrypt_database_setting("database_port")
        if value is None:
            return value
        return int(value)


def encrypt_database_setting(
    target: Workspace,
    value: Optional[Union[str, int]],
    oldvalue: Optional[str],
    initiator,
):
    if value is None:
        return value
    return encrypt(str(value), settings.encryption_key)


event.listen(Workspace.database_host, "set", encrypt_database_setting, retval=True)
event.listen(Workspace.database_port, "set", encrypt_database_setting, retval=True)
event.listen(Workspace.database_username, "set", encrypt_database_setting, retval=True)
event.listen(Workspace.database_password, "set", encrypt_database_setting, retval=True)
event.listen(Workspace.database_name, "set", encrypt_database_setting, retval=True)



"""Add UserField model

Revision ID: b530ce0d8bd9
Revises: bc4021719fac
Create Date: 2022-04-28 15:35:51.333535

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

import fief

# revision identifiers, used by Alembic.
revision = "b530ce0d8bd9"
down_revision = "bc4021719fac"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()
    if connection.dialect.name == "postgresql":
        client_type_enum = postgresql.ENUM(
            "STRING",
            "INTEGER",
            "BOOLEAN",
            "DATE",
            "DATETIME",
            "CHOICE",
            "PHONE_NUMBER",
            "ADDRESS",
            "TIMEZONE",
            name="userfieldtype",
            create_type=False,
        )
        client_type_enum.create(connection, checkfirst=True)
    else:
        client_type_enum = sa.Enum(
            "STRING",
            "INTEGER",
            "BOOLEAN",
            "DATE",
            "DATETIME",
            "CHOICE",
            "PHONE_NUMBER",
            "ADDRESS",
            "TIMEZONE",
            name="userfieldtype",
        )

    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_user_fields",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=320), nullable=False),
        sa.Column("slug", sa.String(length=320), nullable=False),
        sa.Column(
            "type",
            client_type_enum,
            nullable=True,
        ),
        sa.Column("configuration", sa.JSON(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_user_fields_created_at"),
        "fief_user_fields",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_user_fields_slug"), "fief_user_fields", ["slug"], unique=True
    )
    op.create_index(
        op.f("ix_fief_user_fields_type"), "fief_user_fields", ["type"], unique=False
    )
    op.create_index(
        op.f("ix_fief_user_fields_updated_at"),
        "fief_user_fields",
        ["updated_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_fief_user_fields_updated_at"), table_name="fief_user_fields")
    op.drop_index(op.f("ix_fief_user_fields_type"), table_name="fief_user_fields")
    op.drop_index(op.f("ix_fief_user_fields_slug"), table_name="fief_user_fields")
    op.drop_index(op.f("ix_fief_user_fields_created_at"), table_name="fief_user_fields")
    op.drop_table("fief_user_fields")

    try:
        client_type_enum = postgresql.ENUM(
            "STRING",
            "INTEGER",
            "BOOLEAN",
            "DATE",
            "DATETIME",
            "CHOICE",
            "PHONE_NUMBER",
            "ADDRESS",
            "TIMEZONE",
            name="userfieldtype",
            create_type=False,
        )
        client_type_enum.drop(op.get_bind(), checkfirst=True)
    except AttributeError:
        pass
    # ### end Alembic commands ###



from enum import Enum
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, TypedDict

from sqlalchemy import JSON, Column
from sqlalchemy import Enum as SQLEnum
from sqlalchemy import String
from sqlalchemy.orm import relationship

from fief.models.base import WorkspaceBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel

if TYPE_CHECKING:
    from fief.models.user_field_value import UserFieldValue


class UserFieldType(str, Enum):
    STRING = "STRING"
    INTEGER = "INTEGER"
    BOOLEAN = "BOOLEAN"
    DATE = "DATE"
    DATETIME = "DATETIME"
    CHOICE = "CHOICE"
    PHONE_NUMBER = "PHONE_NUMBER"
    ADDRESS = "ADDRESS"
    TIMEZONE = "TIMEZONE"


class UserFieldConfiguration(TypedDict):
    choices: Optional[List[Tuple[str, str]]]
    at_registration: bool
    at_update: bool
    required: bool
    default: Optional[Any]


class UserField(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "user_fields"

    name: str = Column(String(length=320), nullable=False)
    slug: str = Column(String(length=320), index=True, nullable=False, unique=True)
    type: UserFieldType = Column(SQLEnum(UserFieldType), index=True, nullable=True)
    configuration: UserFieldConfiguration = Column(JSON, nullable=False)

    user_field_values: List["UserFieldValue"] = relationship(
        "UserFieldValue", back_populates="user_field", cascade="all, delete"
    )

    def get_required(self) -> bool:
        return self.configuration["required"]

    def get_default(self) -> Optional[Any]:
        return self.configuration["default"]



import binascii

from cryptography import fernet


def generate_key() -> bytes:
    return fernet.Fernet.generate_key()


def is_valid_key(key: bytes) -> bool:
    try:
        fernet.Fernet(key)
    except binascii.Error:
        return False
    else:
        return True


def encrypt(value: str, key: bytes) -> str:
    f = fernet.Fernet(key)
    return f.encrypt(value.encode("utf-8")).decode("utf-8")


def decrypt(value: str, key: bytes) -> str:
    f = fernet.Fernet(key)
    return f.decrypt(value.encode("utf-8")).decode("utf-8")



from typing import Optional, Tuple

from postmarker.core import PostmarkClient
from postmarker.exceptions import ClientError

from fief.services.email.base import EmailProvider, SendEmailError


class Postmark(EmailProvider):
    def __init__(self, server_token: str) -> None:
        self._client = PostmarkClient(server_token=server_token)

    def send_email(
        self,
        *,
        sender: Tuple[str, Optional[str]],
        recipient: Tuple[str, Optional[str]],
        subject: str,
        html: Optional[str] = None,
        text: Optional[str] = None,
    ):
        from_email, from_name = sender
        to_email, to_name = recipient
        try:
            self._client.emails.send(
                From=from_email if from_name is None else f"{from_name} <{from_email}>",
                To=to_email if from_name is None else f"{to_name} <{to_email}>",
                Subject=subject,
                HtmlBody=html,
                TextBody=text,
            )
        except ClientError as e:
            raise SendEmailError(str(e)) from e



from fief.apps.admin.app import app as admin_app
from fief.apps.admin_frontend.app import app as admin_frontend_app
from fief.apps.auth.app import app as auth_app

__all__ = ["admin_app", "admin_frontend_app", "auth_app"]



from typing import List, Optional, Tuple

from fastapi import Depends, HTTPException, Query, status
from sqlalchemy import select

from fief import schemas
from fief.dependencies.pagination import (
    Ordering,
    Pagination,
    get_ordering,
    get_paginated_objects,
    get_pagination,
)
from fief.dependencies.workspace_managers import get_tenant_manager
from fief.errors import APIErrorCode
from fief.managers import TenantManager
from fief.models import Tenant


async def get_current_tenant(
    tenant_slug: Optional[str] = Query(None),
    manager: TenantManager = Depends(get_tenant_manager),
) -> Tenant:
    if tenant_slug is None:
        tenant = await manager.get_default()
    else:
        tenant = await manager.get_by_slug(tenant_slug)

    if tenant is None:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST)

    return tenant


async def get_tenant_from_create_user_internal(
    user_create: schemas.user.UserCreateInternal,
    manager: TenantManager = Depends(get_tenant_manager),
) -> Tenant:
    tenant = await manager.get_by_id(user_create.tenant_id)

    if tenant is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_CREATE_UNKNOWN_TENANT,
        )

    return tenant


async def get_paginated_tenants(
    query: Optional[str] = Query(None),
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(get_ordering),
    manager: TenantManager = Depends(get_tenant_manager),
) -> Tuple[List[Tenant], int]:
    statement = select(Tenant)

    if query is not None:
        statement = statement.where(Tenant.name.ilike(f"%{query}%"))

    return await get_paginated_objects(statement, pagination, ordering, manager)



from typing import List, Tuple

from fastapi import Query
from sqlalchemy.sql import Select

from fief.managers.base import BaseManager, M

PaginatedObjects = Tuple[List[M], int]


async def get_paginated_objects(
    statement: Select,
    pagination: Tuple[int, int],
    ordering: List[Tuple[List[str], bool]],
    manager: BaseManager[M],
) -> PaginatedObjects[M]:
    limit, skip = pagination
    statement = manager.orderize(statement, ordering)
    return await manager.paginate(statement, limit, skip)


Pagination = Tuple[int, int]


async def get_pagination(
    limit: int = Query(10, gt=0), skip: int = Query(0)
) -> Pagination:
    return min(limit, 100), skip


Ordering = List[Tuple[List[str], bool]]


async def get_ordering(ordering: str = Query(None)) -> Ordering:
    ordering_fields = []
    if ordering:
        fields = ordering.split(",")
        for field in fields:
            is_desc = False
            if field.startswith("-"):
                is_desc = True
                field = field[1:]
            ordering_fields.append((field.split("."), is_desc))
    return ordering_fields



from typing import Optional

from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import LoginSession


class LoginSessionManager(BaseManager[LoginSession], UUIDManagerMixin[LoginSession]):
    model = LoginSession

    async def get_by_token(self, token: str) -> Optional[LoginSession]:
        statement = select(LoginSession).where(LoginSession.token == token)
        return await self.get_one_or_none(statement)



import re
from typing import List, Optional

from pydantic import UUID4

from fief.db.workspace import get_workspace_session
from fief.managers import (
    ClientManager,
    TenantManager,
    WorkspaceManager,
    WorkspaceUserManager,
)
from fief.models import Client, Tenant, Workspace, WorkspaceUser
from fief.schemas.workspace import WorkspaceCreate
from fief.services.main_workspace import get_main_fief_client, get_main_fief_workspace
from fief.services.workspace_db import (
    WorkspaceDatabase,
    WorkspaceDatabaseConnectionError,
)

LOCALHOST_HOST_PATTERN = re.compile(r"([^\.]+\.)?localhost(\d+)?", flags=re.IGNORECASE)


class WorkspaceCreation:
    def __init__(
        self,
        workspace_manager: WorkspaceManager,
        workspace_user_manager: WorkspaceUserManager,
        workspace_db: WorkspaceDatabase,
    ) -> None:
        self.workspace_manager = workspace_manager
        self.workspace_user_manager = workspace_user_manager
        self.workspace_db = workspace_db

    async def create(
        self,
        workspace_create: WorkspaceCreate,
        user_id: Optional[UUID4] = None,
        default_domain: Optional[str] = None,
        default_client_id: Optional[str] = None,
        default_client_secret: Optional[str] = None,
        default_redirect_uris: Optional[List[str]] = None,
        default_encryption_key: Optional[str] = None,
    ) -> Workspace:
        workspace = Workspace(**workspace_create.dict())

        if default_domain is None:
            # Create workspace on available subdomain
            domain = await self.workspace_manager.get_available_subdomain(
                workspace.name
            )
            workspace.domain = domain
        else:
            workspace.domain = default_domain

        workspace = await self.workspace_manager.create(workspace)

        # Apply the database schema
        try:
            alembic_revision = self.workspace_db.migrate(
                workspace.get_database_url(False), workspace.get_schema_name()
            )
        except WorkspaceDatabaseConnectionError:
            await self.workspace_manager.delete(workspace)
            raise

        workspace.alembic_revision = alembic_revision
        await self.workspace_manager.update(workspace)

        # Link the user to this workspace
        if user_id is not None:
            workspace_user = WorkspaceUser(workspace_id=workspace.id, user_id=user_id)
            await self.workspace_user_manager.create(workspace_user)

        # Create a default tenant and client
        async with get_workspace_session(workspace) as session:
            tenant_name = workspace.name
            tenant_slug = await TenantManager(session).get_available_slug(tenant_name)
            tenant = Tenant(name=workspace.name, slug=tenant_slug, default=True)

            session.add(tenant)

            client = Client(
                name=f"{tenant.name}'s client",
                first_party=True,
                tenant=tenant,
                redirect_uris=default_redirect_uris
                if default_redirect_uris is not None
                else ["http://localhost:8000/docs/oauth2-redirect"],
            )

            if default_client_id is not None:
                client.client_id = default_client_id
            if default_client_secret is not None:
                client.client_secret = default_client_secret
            if default_encryption_key is not None:
                client.encrypt_jwk = default_encryption_key

            session.add(client)

            await session.commit()

        # Allow a redirect URI on this workspace on the main Fief client
        await self._add_fief_redirect_uri(workspace)

        return workspace

    async def _add_fief_redirect_uri(self, workspace: Workspace):
        main_workspace = await get_main_fief_workspace()

        async with get_workspace_session(main_workspace) as session:
            client_manager = ClientManager(session)
            fief_client = await get_main_fief_client()

            localhost_domain = LOCALHOST_HOST_PATTERN.match(workspace.domain)
            redirect_uri = f"{'http' if localhost_domain else 'https'}://{workspace.domain}/admin/api/auth/callback"
            fief_client.redirect_uris = fief_client.redirect_uris + [redirect_uri]

            await client_manager.update(fief_client)



from gettext import gettext as _
from typing import Any, Dict, List, Optional, Union

from pydantic import ValidationError
from starlette.requests import FormData

from fief.locale import Translations
from fief.models import Client, Tenant
from fief.schemas.auth import (
    AuthorizeError,
    AuthorizeRedirectError,
    ConsentError,
    LoginError,
    LogoutError,
    TokenError,
)
from fief.schemas.register import RegisterError
from fief.schemas.reset import ResetPasswordError

PYDANTIC_ERROR_MESSAGES = {
    "value_error.missing": _("This field is required."),
    "value_error.any_str.min_length": _("This field is required."),
    "value_error.email": _("This email address is invalid."),
    "value_error.date": _("This date is invalid."),
    "value_error.datetime": _("This date and time is invalid."),
    "value_error.phone_number.invalid": _("This phone number is invalid."),
    "value_error.phone_number.missing_region": _("The country code is missing."),
    "value_error.country_code.invalid": _("This country code is invalid."),
    "value_error.timezone.invalid": _("This timezone is invalid."),
    "value_error.boolean.must_be_true": _("This must be checked."),
    "type_error.bool": _("This value is invalid."),
    "type_error.integer": _("This value is invalid."),
    "type_error.enum": _("This value is invalid."),
}


class FormValidationError(ValidationError):
    def __init__(
        self,
        template: str,
        context: Dict[str, Any],
        translations: Translations,
        *args,
        **kwargs,
    ) -> None:
        self.template = template
        self.context = context
        self.translations = translations
        super().__init__(*args, **kwargs)

    def form_errors(self) -> Dict[Union[str, int], Any]:
        form_errors: Dict[Union[str, int], Any] = {}
        for error in self.errors():
            error_dict = form_errors
            for key in error["loc"][:-1]:
                error_dict = error_dict.setdefault(key, {})
            error_dict[error["loc"][-1]] = self.translations.gettext(
                PYDANTIC_ERROR_MESSAGES[error["type"]]
            )
        return form_errors


class RegisterException(Exception):
    def __init__(
        self,
        error: RegisterError,
        context: Dict[str, Any],
        form_data: Optional[Dict[str, Any]] = None,
        *,
        fatal: bool = False,
    ) -> None:
        self.error = error
        self.context = context
        self.form_data = form_data
        self.fatal = fatal


class AuthorizeException(Exception):
    def __init__(self, error: AuthorizeError, tenant: Optional[Tenant] = None) -> None:
        self.error = error
        self.tenant = tenant


class AuthorizeRedirectException(Exception):
    def __init__(
        self,
        error: AuthorizeRedirectError,
        redirect_uri: str,
        response_mode: str,
        state: Optional[str],
        tenant: Optional[Tenant] = None,
    ) -> None:
        self.error = error
        self.redirect_uri = redirect_uri
        self.response_mode = response_mode
        self.state = state
        self.tenant = tenant


class LoginException(Exception):
    def __init__(
        self, error: LoginError, tenant: Optional[Tenant] = None, *, fatal: bool = False
    ) -> None:
        self.error = error
        self.tenant = tenant
        self.fatal = fatal


class ConsentException(Exception):
    def __init__(
        self,
        error: ConsentError,
        client: Client,
        scope: List[str],
        tenant: Optional[Tenant] = None,
        *,
        fatal: bool = False,
    ) -> None:
        self.error = error
        self.client = client
        self.scope = scope
        self.tenant = tenant
        self.fatal = fatal


class TokenRequestException(Exception):
    def __init__(self, error: TokenError) -> None:
        self.error = error


class ResetPasswordException(Exception):
    def __init__(
        self,
        error: ResetPasswordError,
        form_data: Optional[FormData] = None,
        tenant: Optional[Tenant] = None,
        *,
        fatal: bool = False,
    ) -> None:
        self.error = error
        self.form_data = form_data
        self.tenant = tenant
        self.fatal = fatal


class LogoutException(Exception):
    def __init__(self, error: LogoutError, tenant: Optional[Tenant] = None) -> None:
        self.error = error
        self.tenant = tenant



from typing import TYPE_CHECKING

from sqlalchemy import engine, event
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.orm import sessionmaker

from fief.settings import settings

if TYPE_CHECKING:
    from fief.models import Workspace


def create_engine(database_url: engine.URL) -> AsyncEngine:
    engine = create_async_engine(
        database_url,
        echo=settings.log_level == "DEBUG",
        pool_recycle=settings.database_pool_recycle_seconds,
    )
    dialect_name = engine.dialect.name

    # Special tweak for SQLite to better handle transaction
    # See: https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#serializable-isolation-savepoints-transactional-ddl
    if dialect_name == "sqlite":

        @event.listens_for(engine.sync_engine, "connect")
        def do_connect(dbapi_connection, connection_record):
            # disable pysqlite's emitting of the BEGIN statement entirely.
            # also stops it from emitting COMMIT before any DDL.
            dbapi_connection.isolation_level = None

        @event.listens_for(engine.sync_engine, "begin")
        def do_begin(conn):
            # emit our own BEGIN
            conn.exec_driver_sql("BEGIN")

    return engine


def create_async_session_maker(engine: AsyncEngine):
    return sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)


__all__ = [
    "create_engine",
    "create_async_session_maker",
]



import asyncio
import contextlib
import json
import uuid
from typing import Any, AsyncGenerator, Dict, Optional, Tuple
from unittest.mock import MagicMock

import asgi_lifespan
import httpx
import pytest
from fastapi import FastAPI
from sqlalchemy import engine
from sqlalchemy_utils import create_database, drop_database

from fief.apps import admin_app, auth_app
from fief.crypto.access_token import generate_access_token
from fief.crypto.token import generate_token
from fief.csrf import check_csrf
from fief.db import AsyncConnection, AsyncEngine, AsyncSession
from fief.db.engine import create_engine
from fief.db.main import get_main_async_session
from fief.db.types import DatabaseType, get_driver
from fief.db.workspace import get_connection
from fief.dependencies.current_workspace import get_current_workspace_session
from fief.dependencies.fief import FiefAsyncRelativeEndpoints, get_fief
from fief.dependencies.tasks import get_send_task
from fief.dependencies.workspace_creation import get_workspace_creation
from fief.dependencies.workspace_db import get_workspace_db
from fief.models import (
    AdminAPIKey,
    AdminSessionToken,
    MainBase,
    User,
    Workspace,
    WorkspaceUser,
)
from fief.services.workspace_creation import WorkspaceCreation
from fief.services.workspace_db import WorkspaceDatabase
from fief.settings import settings
from fief.tasks import send_task
from tests.data import TestData, data_mapping, session_token_tokens
from tests.types import GetTestDatabase, TenantParams, TestClientGeneratorType

pytest.register_assert_rewrite("tests.helpers")


@pytest.fixture(scope="session")
def event_loop():
    """Force the pytest-asyncio loop to be the main one."""
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def get_test_database() -> GetTestDatabase:
    @contextlib.asynccontextmanager
    async def _get_test_database(
        *, name: str = "fief-test"
    ) -> AsyncGenerator[Tuple[engine.URL, DatabaseType], None]:
        url = settings.get_database_url(False).set(database=name)
        assert url.database == name
        create_database(url)
        yield (url, settings.database_type)
        drop_database(url)

    return _get_test_database


@pytest.fixture(scope="session")
async def main_test_database(
    get_test_database: GetTestDatabase,
) -> AsyncGenerator[Tuple[engine.URL, DatabaseType], None]:
    async with get_test_database() as (url, database_type):
        url = url.set(drivername=get_driver(database_type, asyncio=True))
        yield url, database_type


@pytest.fixture(scope="session")
async def main_engine(
    main_test_database: Tuple[engine.URL, DatabaseType],
) -> AsyncGenerator[AsyncEngine, None]:
    url, _ = main_test_database
    engine = create_engine(url)
    yield engine
    await engine.dispose()


@pytest.fixture(scope="session")
async def main_connection(
    main_engine: AsyncEngine,
) -> AsyncGenerator[AsyncConnection, None]:
    async with main_engine.connect() as connection:
        yield connection


@pytest.fixture(scope="session")
async def create_main_db(main_connection: AsyncConnection):
    await main_connection.run_sync(MainBase.metadata.create_all)


@pytest.fixture(scope="session")
async def main_session(
    main_connection: AsyncConnection, create_main_db
) -> AsyncGenerator[AsyncSession, None]:
    async with AsyncSession(bind=main_connection, expire_on_commit=False) as session:
        await session.begin_nested()
        yield session
        await session.rollback()


@pytest.fixture(scope="session")
def main_session_manager(main_session: AsyncSession):
    @contextlib.asynccontextmanager
    async def _main_session_manager(*args, **kwargs):
        yield main_session

    return _main_session_manager


@pytest.fixture(scope="session", autouse=True)
@pytest.mark.asyncio
async def workspace(
    main_test_database: Tuple[engine.URL, DatabaseType],
    main_session,
    create_main_db,
) -> AsyncGenerator[Workspace, None]:
    url, database_type = main_test_database
    workspace = Workspace(
        name="Duché de Bretagne",
        domain="bretagne.localhost:8000",
        database_type=database_type,
        database_host=url.host,
        database_port=url.port,
        database_username=url.username,
        database_password=url.password,
        database_name=url.database,
        alembic_revision="LATEST",
    )
    main_session.add(workspace)
    await main_session.commit()

    workspace_db = WorkspaceDatabase()
    workspace_db.migrate(workspace.get_database_url(False), workspace.get_schema_name())

    yield workspace


@pytest.fixture(scope="session")
async def workspace_engine(workspace: Workspace) -> AsyncGenerator[AsyncEngine, None]:
    engine = create_engine(workspace.get_database_url())
    yield engine
    await engine.dispose()


@pytest.fixture(scope="session")
async def workspace_connection(
    workspace_engine: AsyncEngine, workspace: Workspace
) -> AsyncGenerator[AsyncConnection, None]:
    async with get_connection(
        workspace_engine, workspace.get_schema_name()
    ) as connection:
        await connection.begin()
        yield connection
        await connection.rollback()


@pytest.fixture(scope="session")
@pytest.mark.asyncio
async def test_data(workspace_connection: AsyncConnection) -> TestData:
    async with AsyncSession(
        bind=workspace_connection, expire_on_commit=False
    ) as session:
        for model in data_mapping.values():
            for object in model.values():
                session.add(object)
        await session.commit()
    await workspace_connection.commit()
    yield data_mapping


@pytest.fixture
async def workspace_session(
    workspace_connection: AsyncConnection,
) -> AsyncGenerator[AsyncSession, None]:
    await workspace_connection.begin_nested()
    async with AsyncSession(
        bind=workspace_connection, expire_on_commit=False
    ) as session:
        yield session
    await workspace_connection.rollback()


@pytest.fixture
def workspace_session_manager(workspace_session: AsyncSession):
    @contextlib.asynccontextmanager
    async def _workspace_session_manager(*args, **kwargs):
        yield workspace_session

    return _workspace_session_manager


@pytest.fixture
def not_existing_uuid() -> uuid.UUID:
    return uuid.uuid4()


@pytest.fixture
async def workspace_db_mock() -> MagicMock:
    mock = MagicMock(spec=WorkspaceDatabase)
    mock.get_latest_revision.return_value = "LATEST"
    return mock


@pytest.fixture
async def workspace_creation_mock() -> MagicMock:
    return MagicMock(spec=WorkspaceCreation)


@pytest.fixture
async def fief_client_mock() -> MagicMock:
    return MagicMock(spec=FiefAsyncRelativeEndpoints)


@pytest.fixture
async def send_task_mock() -> MagicMock:
    return MagicMock(spec=send_task)


@pytest.fixture
def workspace_host(
    request: pytest.FixtureRequest, workspace: Workspace
) -> Optional[str]:
    marker = request.node.get_closest_marker("workspace_host")
    if marker:
        return workspace.domain
    return None


@pytest.fixture
def workspace_admin_user() -> User:
    return User(
        id=uuid.uuid4(),
        email="dev@bretagne.duchy",
        hashed_password="dev",
        tenant_id=uuid.uuid4(),
    )


@pytest.fixture
async def admin_session_token(
    main_session: AsyncSession,
    workspace: Workspace,
    workspace_admin_user: User,
) -> AsyncGenerator[Tuple[AdminSessionToken, str], None]:
    workspace_user = WorkspaceUser(
        workspace_id=workspace.id, user_id=workspace_admin_user.id
    )
    main_session.add(workspace_user)
    await main_session.commit()

    token, token_hash = generate_token()
    session_token = AdminSessionToken(
        token=token_hash,
        raw_tokens="{}",
        raw_userinfo=json.dumps(workspace_admin_user.get_claims()),
    )
    main_session.add(session_token)

    await main_session.commit()

    yield (session_token, token)

    await main_session.delete(session_token)
    await main_session.delete(workspace_user)


@pytest.fixture
async def admin_api_key(
    main_session: AsyncSession, workspace: Workspace
) -> AsyncGenerator[Tuple[AdminAPIKey, str], None]:
    token, token_hash = generate_token()
    admin_api_key = AdminAPIKey(
        name="API Key", token=token_hash, workspace_id=workspace.id
    )
    main_session.add(admin_api_key)
    await main_session.commit()

    yield (admin_api_key, token)

    await main_session.delete(admin_api_key)


@pytest.fixture
async def authenticated_admin(
    request: pytest.FixtureRequest,
    admin_session_token: Tuple[AdminSessionToken, str],
    admin_api_key: Tuple[AdminAPIKey, str],
) -> Dict[str, Any]:
    marker = request.node.get_closest_marker("authenticated_admin")
    headers = {}
    if marker:
        mode = marker.kwargs.get("mode", "api_key")
        if mode == "session":
            headers[
                "Cookie"
            ] = f"{settings.fief_admin_session_cookie_name}={admin_session_token[1]}"
        elif mode == "api_key":
            headers["Authorization"] = f"Bearer {admin_api_key[1]}"
    return headers


@pytest.fixture(
    params=[
        {
            "path_prefix": "",
            "tenant_alias": "default",
            "client_alias": "default_tenant",
            "user_alias": "regular",
            "login_session_alias": "default",
            "session_token_alias": "regular",
        },
        {
            "path_prefix": "/secondary",
            "tenant_alias": "secondary",
            "client_alias": "secondary_tenant",
            "user_alias": "regular_secondary",
            "login_session_alias": "secondary",
            "session_token_alias": "regular_secondary",
        },
    ]
)
def tenant_params(request, test_data: TestData) -> TenantParams:
    params = request.param
    return TenantParams(
        path_prefix=params["path_prefix"],
        tenant=test_data["tenants"][params["tenant_alias"]],
        client=test_data["clients"][params["client_alias"]],
        user=test_data["users"][params["user_alias"]],
        login_session=test_data["login_sessions"][params["login_session_alias"]],
        session_token=test_data["session_tokens"][params["session_token_alias"]],
        session_token_token=session_token_tokens[params["session_token_alias"]],
    )


@pytest.fixture
def access_token(
    request: pytest.FixtureRequest,
    test_data: TestData,
    tenant_params: TenantParams,
    workspace: Workspace,
) -> Optional[str]:
    marker = request.node.get_closest_marker("access_token")
    if marker:
        from_tenant_params: bool = marker.kwargs.get("from_tenant_params", False)
        if from_tenant_params:
            user = tenant_params.user
        else:
            user_alias = marker.kwargs["user"]
            user = test_data["users"][user_alias]

        user_tenant = user.tenant
        client = next(
            client
            for _, client in test_data["clients"].items()
            if client.tenant_id == user_tenant.id
        )

        return generate_access_token(
            user_tenant.get_sign_jwk(),
            user_tenant.get_host(workspace.domain),
            client,
            user,
            ["openid"],
            3600,
        )
    return None


@pytest.fixture
async def test_client_admin_generator(
    main_session: AsyncSession,
    workspace_session: AsyncSession,
    workspace_db_mock: MagicMock,
    workspace_creation_mock: MagicMock,
    send_task_mock: MagicMock,
    fief_client_mock: MagicMock,
    authenticated_admin: Dict[str, Any],
    workspace_host: Optional[str],
) -> TestClientGeneratorType:
    @contextlib.asynccontextmanager
    async def _test_client_generator(app: FastAPI):
        app.dependency_overrides = {}
        app.dependency_overrides[get_main_async_session] = lambda: main_session
        app.dependency_overrides[
            get_current_workspace_session
        ] = lambda: workspace_session
        app.dependency_overrides[get_workspace_db] = lambda: workspace_db_mock
        app.dependency_overrides[
            get_workspace_creation
        ] = lambda: workspace_creation_mock
        app.dependency_overrides[get_send_task] = lambda: send_task_mock
        app.dependency_overrides[get_fief] = lambda: fief_client_mock
        settings.fief_admin_session_cookie_domain = ""

        headers = {**authenticated_admin}
        if workspace_host is not None:
            headers["Host"] = workspace_host

        async with asgi_lifespan.LifespanManager(app):
            async with httpx.AsyncClient(
                app=app,
                base_url="http://api.fief.dev",
                headers=headers,
            ) as test_client:
                yield test_client

    return _test_client_generator


@pytest.fixture
async def test_client_admin(
    test_client_admin_generator: TestClientGeneratorType,
) -> AsyncGenerator[httpx.AsyncClient, None]:
    async with test_client_admin_generator(admin_app) as test_client:
        yield test_client


@pytest.fixture
async def test_client_auth_generator(
    main_session: AsyncSession,
    workspace_session: AsyncSession,
    workspace_db_mock: MagicMock,
    workspace_creation_mock: MagicMock,
    send_task_mock: MagicMock,
    workspace_host: Optional[str],
    access_token: Optional[str],
) -> TestClientGeneratorType:
    @contextlib.asynccontextmanager
    async def _test_client_generator(app: FastAPI):
        app.dependency_overrides = {}
        app.dependency_overrides[get_main_async_session] = lambda: main_session
        app.dependency_overrides[
            get_current_workspace_session
        ] = lambda: workspace_session
        app.dependency_overrides[get_workspace_db] = lambda: workspace_db_mock
        app.dependency_overrides[
            get_workspace_creation
        ] = lambda: workspace_creation_mock
        app.dependency_overrides[check_csrf] = lambda: None
        app.dependency_overrides[get_send_task] = lambda: send_task_mock

        headers = {}
        if workspace_host is not None:
            headers["Host"] = workspace_host
        if access_token is not None:
            headers["Authorization"] = f"Bearer {access_token}"

        async with asgi_lifespan.LifespanManager(app):
            async with httpx.AsyncClient(
                app=app, base_url="http://api.fief.dev", headers=headers
            ) as test_client:
                yield test_client

    return _test_client_generator


@pytest.fixture
async def test_client_auth(
    test_client_auth_generator: TestClientGeneratorType,
) -> AsyncGenerator[httpx.AsyncClient, None]:
    async with test_client_auth_generator(auth_app) as test_client:
        yield test_client



from typing import AsyncGenerator, Optional

from fastapi import Header, HTTPException, status
from fastapi.param_functions import Depends

from fief.db import AsyncSession
from fief.db.workspace import get_workspace_session
from fief.dependencies.main_managers import get_workspace_manager
from fief.dependencies.workspace_db import get_workspace_db
from fief.errors import APIErrorCode
from fief.managers import WorkspaceManager
from fief.models import Workspace
from fief.services.workspace_db import WorkspaceDatabase


async def get_host(
    host: Optional[str] = Header(None, include_in_schema=False)
) -> Optional[str]:
    return host


async def get_current_workspace(
    host: Optional[str] = Depends(get_host),
    manager: WorkspaceManager = Depends(get_workspace_manager),
    workspace_db: WorkspaceDatabase = Depends(get_workspace_db),
) -> Workspace:
    workspace = None
    if host is not None:
        workspace = await manager.get_by_domain(host)

    if workspace is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.CANT_DETERMINE_VALID_WORKSPACE,
        )

    latest_revision = workspace_db.get_latest_revision()
    if workspace.alembic_revision != latest_revision:
        raise HTTPException(
            status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
            detail=APIErrorCode.WORKSPACE_DB_OUTDATED_MIGRATION,
        )

    return workspace


async def get_current_workspace_session(
    workspace: Workspace = Depends(get_current_workspace),
) -> AsyncGenerator[AsyncSession, None]:
    try:
        async with get_workspace_session(workspace) as session:
            yield session
    except ConnectionError as e:
        raise HTTPException(
            status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
            detail=APIErrorCode.WORKSPACE_DB_CONNECTION_ERROR,
        ) from e



from datetime import datetime, timedelta, timezone
from typing import Optional

from sqlalchemy import select

from fief.managers.base import BaseManager, UUIDManagerMixin
from fief.models import AuthorizationCode
from fief.settings import settings


class AuthorizationCodeManager(
    BaseManager[AuthorizationCode], UUIDManagerMixin[AuthorizationCode]
):
    model = AuthorizationCode

    async def get_valid_by_code(self, code: str) -> Optional[AuthorizationCode]:
        max_created_at = datetime.now(timezone.utc) - timedelta(
            seconds=settings.authorization_code_lifetime_seconds
        )
        statement = select(AuthorizationCode).where(
            AuthorizationCode.code == code,
            AuthorizationCode.created_at > max_created_at,
        )
        return await self.get_one_or_none(statement)

    async def get_by_code(self, code: str) -> Optional[AuthorizationCode]:
        statement = select(AuthorizationCode).where(AuthorizationCode.code == code)
        return await self.get_one_or_none(statement)



from datetime import date, datetime
from typing import TYPE_CHECKING, Any, Tuple

from pydantic import UUID4
from sqlalchemy import JSON, Boolean, Column, Date, ForeignKey, Integer, Text
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import UniqueConstraint

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, TIMESTAMPAware, UUIDModel
from fief.models.user import User
from fief.models.user_field import UserField, UserFieldType

if TYPE_CHECKING:
    hybrid_property = property
else:
    from sqlalchemy.ext.hybrid import hybrid_property


class UserFieldValue(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "user_field_values"
    __table_args__ = (UniqueConstraint("user_field_id", "user_id"),)

    value_string = Column(Text, nullable=True)
    value_integer = Column(Integer, nullable=True)
    value_boolean = Column(Boolean, nullable=True)
    value_date = Column(Date, nullable=True)
    value_datetime = Column(TIMESTAMPAware(timezone=True), nullable=True)
    value_json = Column(JSON, nullable=True)

    user_id: UUID4 = Column(GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    user: User = relationship("User", back_populates="user_field_values")

    user_field_id: UUID4 = Column(GUID, ForeignKey(UserField.id, ondelete="CASCADE"), nullable=False)  # type: ignore
    user_field: UserField = relationship(
        "UserField", back_populates="user_field_values", lazy="selectin"
    )

    def _get_field_value(self) -> str:
        user_field_type = self.user_field.type
        if user_field_type in [UserFieldType.INTEGER]:
            return "value_integer"
        elif user_field_type in [UserFieldType.BOOLEAN]:
            return "value_boolean"
        elif user_field_type in [UserFieldType.DATE]:
            return "value_date"
        elif user_field_type in [UserFieldType.DATETIME]:
            return "value_datetime"
        elif user_field_type in [UserFieldType.ADDRESS]:
            return "value_json"
        else:
            return "value_string"

    @hybrid_property
    def value(self) -> Any:
        field_value = self._get_field_value()
        return getattr(self, field_value)

    @value.setter
    def value(self, value: Any):
        field_value = self._get_field_value()
        setattr(self, field_value, value)

    def get_slug_and_value(self, *, json_serializable: bool = False) -> Tuple[str, Any]:
        value = self.value
        if json_serializable and (
            isinstance(value, date) or isinstance(value, datetime)
        ):
            value = value.isoformat()
        return self.user_field.slug, value



import json
import uuid
from datetime import datetime, timezone
from typing import List

from jwcrypto import jwk, jwt
from jwcrypto.common import JWException
from pydantic import UUID4

from fief.models import Client, User


class InvalidAccessToken(Exception):
    pass


def generate_access_token(
    key: jwk.JWK,
    host: str,
    client: Client,
    user: User,
    scope: List[str],
    lifetime_seconds: int,
) -> str:
    iat = int(datetime.now(timezone.utc).timestamp())
    exp = iat + lifetime_seconds

    claims = {
        "iss": host,
        "sub": str(user.id),
        "aud": [client.client_id],
        "exp": exp,
        "iat": iat,
        "azp": client.client_id,
        "scope": " ".join(scope),
    }

    token = jwt.JWT(header={"alg": "RS256"}, claims=claims)
    token.make_signed_token(key)

    return token.serialize()


def read_access_token(key: jwk.JWK, token: str) -> UUID4:
    try:
        decoded_jwt = jwt.JWT(jwt=token, key=key)
        claims = json.loads(decoded_jwt.claims)
        user_id = claims["sub"]
    except (JWException, KeyError, ValueError) as e:
        raise InvalidAccessToken() from e
    else:
        return uuid.UUID(user_id)
